@@ -15,8 +15,6 @@ module GHC.ByteCode.Asm (
         mkTupleInfoLit
   ) where
 
-#include "HsVersions.h"
-
 import GHC.Prelude
 
 import GHC.ByteCode.Instr
@@ -34,7 +32,7 @@ import GHC.Types.Unique.DSet
 
 import GHC.Utils.Outputable
 import GHC.Utils.Panic
-import GHC.Utils.Misc
+import GHC.Utils.Panic.Plain
 
 import GHC.Core.TyCon
 import GHC.Data.FastString
@@ -204,7 +202,7 @@ assembleBCO platform (ProtoBCO { protoBCOName       = nm
   (final_insns, final_lits, final_ptrs) <- flip execStateT initial_state $ runAsm platform long_jumps env asm
 
   -- precomputed size should be equal to final size
-  ASSERT(n_insns == sizeSS final_insns) return ()
+  massert (n_insns == sizeSS final_insns)
 
   let asm_insns = ssElts final_insns
       insns_arr = Array.listArray (0, fromIntegral n_insns - 1) asm_insns
@@ -342,7 +340,7 @@ inspectAsm platform long_jumps initial_offset
 --      count (LargeOp _) = largeArg16s platform
 
 -- Bring in all the bci_ bytecode constants.
-#include "rts/Bytecodes.h"
+#include "Bytecodes.h"
 
 largeArgInstr :: Word16 -> Word16
 largeArgInstr bci = bci_FLAG_LARGE_ARGS .|. bci
@@ -490,8 +488,7 @@ assembleI platform i = case i of
       LitNumWord32  -> int32 (fromIntegral i)
       LitNumInt64   -> int64 (fromIntegral i)
       LitNumWord64  -> int64 (fromIntegral i)
-      LitNumInteger -> panic "GHC.ByteCode.Asm.literal: LitNumInteger"
-      LitNumNatural -> panic "GHC.ByteCode.Asm.literal: LitNumNatural"
+      LitNumBigNat  -> panic "GHC.ByteCode.Asm.literal: LitNumBigNat"
 
     -- We can lower 'LitRubbish' to an arbitrary constant, but @NULL@ is most
     -- likely to elicit a crash (rather than corrupt memory) in case absence
@@ -500,7 +497,7 @@ assembleI platform i = case i of
 
     litlabel fs = lit [BCONPtrLbl fs]
     addr (RemotePtr a) = words [fromIntegral a]
-    float = words . mkLitF
+    float = words . mkLitF platform
     double = words . mkLitD platform
     int = words . mkLitI
     int8 = words . mkLitI64 platform
@@ -569,9 +566,9 @@ mkTupleInfoSig platform TupleInfo{..}
               text "Use -fobject-code to get around this limit"
              )
   | otherwise
-  = ASSERT(length regs <= 24) {- 24 bits for bitmap -}
-    ASSERT(tupleNativeStackSize < 255) {- 8 bits for stack size -}
-    ASSERT(all (`elem` regs) (regSetToList tupleRegs)) {- all regs accounted for -}
+  = assert (length regs <= 24) {- 24 bits for bitmap -}
+    assert (tupleNativeStackSize < 255) {- 8 bits for stack size -}
+    assert (all (`elem` regs) (regSetToList tupleRegs)) {- all regs accounted for -}
     foldl' reg_bit 0 (zip regs [0..]) .|.
       (fromIntegral tupleNativeStackSize `shiftL` 24)
   where
@@ -589,18 +586,28 @@ mkTupleInfoLit platform tuple_info =
 -- words are placed in memory at increasing addresses, the
 -- bit pattern is correct for the host's word size and endianness.
 mkLitI   ::             Int    -> [Word]
-mkLitF   ::             Float  -> [Word]
+mkLitF   :: Platform -> Float  -> [Word]
 mkLitD   :: Platform -> Double -> [Word]
 mkLitI64 :: Platform -> Int64  -> [Word]
 
-mkLitF f
-   = runST (do
+mkLitF platform f = case platformWordSize platform of
+  PW4 -> runST $ do
         arr <- newArray_ ((0::Int),0)
         writeArray arr 0 f
         f_arr <- castSTUArray arr
         w0 <- readArray f_arr 0
         return [w0 :: Word]
-     )
+
+  PW8 -> runST $ do
+        arr <- newArray_ ((0::Int),1)
+        writeArray arr 0 f
+        -- on 64-bit architectures we read two (32-bit) Float cells when we read
+        -- a (64-bit) Word: so we write a dummy value in the second cell to
+        -- avoid an out-of-bound read.
+        writeArray arr 1 0.0
+        f_arr <- castSTUArray arr
+        w0 <- readArray f_arr 0
+        return [w0 :: Word]
 
 mkLitD platform d = case platformWordSize platform of
    PW4 -> runST (do
