@@ -12,51 +12,38 @@ case, so that we don't allocate things, save them on the stack, and
 then discover that they aren't needed in the chosen branch.
 -}
 
-{-# LANGUAGE CPP #-}
+
 {-# OPTIONS_GHC -fprof-auto #-}
 {-# OPTIONS_GHC -Wno-incomplete-uni-patterns #-}
 
 module GHC.Core.Opt.FloatIn ( floatInwards ) where
 
-#include "HsVersions.h"
-
 import GHC.Prelude
 import GHC.Platform
 
-import GHC.Driver.Session
-
 import GHC.Core
 import GHC.Core.Make hiding ( wrapFloats )
 import GHC.Core.Utils
 import GHC.Core.FVs
-import GHC.Core.Opt.Monad   ( CoreM )
 import GHC.Core.Type
 
-import GHC.Types.Basic      ( RecFlag(..), isRec )
+import GHC.Types.Basic      ( RecFlag(..), isRec, Levity(Unlifted) )
 import GHC.Types.Id         ( isOneShotBndr, idType, isJoinId, isJoinId_maybe )
 import GHC.Types.Tickish
 import GHC.Types.Var
 import GHC.Types.Var.Set
 
-import GHC.Unit.Module.ModGuts
-
 import GHC.Utils.Misc
 import GHC.Utils.Panic
-
-import GHC.Utils.Outputable
-
-import Data.List        ( mapAccumL )
+import GHC.Utils.Panic.Plain
 
 {-
 Top-level interface function, @floatInwards@.  Note that we do not
 actually float any bindings downwards from the top-level.
 -}
 
-floatInwards :: ModGuts -> CoreM ModGuts
-floatInwards pgm@(ModGuts { mg_binds = binds })
-  = do { dflags <- getDynFlags
-       ; let platform = targetPlatform dflags
-       ; return (pgm { mg_binds = map (fi_top_bind platform) binds }) }
+floatInwards :: Platform -> CoreProgram -> CoreProgram
+floatInwards platform binds = map (fi_top_bind platform) binds
   where
     fi_top_bind platform (NonRec binder rhs)
       = NonRec binder (fiExpr platform [] (freeVars rhs))
@@ -136,7 +123,7 @@ the closure for a is not built.
 ************************************************************************
 -}
 
-type FreeVarSet  = DVarSet
+type FreeVarSet  = DIdSet
 type BoundVarSet = DIdSet
 
 data FloatInBind = FB BoundVarSet FreeVarSet FloatBind
@@ -144,33 +131,28 @@ data FloatInBind = FB BoundVarSet FreeVarSet FloatBind
         -- of recursive bindings, the set doesn't include the bound
         -- variables.
 
-type FloatInBinds    = [FloatInBind] -- In normal dependency order
-                                     --    (outermost binder first)
-type RevFloatInBinds = [FloatInBind] -- In reverse dependency order
-                                     --    (innermost binder first)
-
-instance Outputable FloatInBind where
-  ppr (FB bvs fvs _) = text "FB" <> braces (sep [ text "bndrs =" <+> ppr bvs
-                                                , text "fvs =" <+> ppr fvs ])
+type FloatInBinds = [FloatInBind]
+        -- In reverse dependency order (innermost binder first)
 
 fiExpr :: Platform
-       -> RevFloatInBinds   -- Binds we're trying to drop
+       -> FloatInBinds      -- Binds we're trying to drop
                             -- as far "inwards" as possible
        -> CoreExprWithFVs   -- Input expr
        -> CoreExpr          -- Result
 
 fiExpr _ to_drop (_, AnnLit lit)     = wrapFloats to_drop (Lit lit)
                                        -- See Note [Dead bindings]
-fiExpr _ to_drop (_, AnnType ty)     = ASSERT( null to_drop ) Type ty
+fiExpr _ to_drop (_, AnnType ty)     = assert (null to_drop) $ Type ty
 fiExpr _ to_drop (_, AnnVar v)       = wrapFloats to_drop (Var v)
 fiExpr _ to_drop (_, AnnCoercion co) = wrapFloats to_drop (Coercion co)
 fiExpr platform to_drop (_, AnnCast expr (co_ann, co))
-  = wrapFloats drop_here $
+  = wrapFloats (drop_here ++ co_drop) $
     Cast (fiExpr platform e_drop expr) co
   where
-    (drop_here, [e_drop])
-      = sepBindsByDropPoint platform False to_drop
-          (freeVarsOfAnn co_ann) [freeVarsOf expr]
+    [drop_here, e_drop, co_drop]
+      = sepBindsByDropPoint platform False
+          [freeVarsOf expr, freeVarsOfAnn co_ann]
+          to_drop
 
 {-
 Applications: we do float inside applications, mainly because we
@@ -179,7 +161,7 @@ pull out any silly ones.
 -}
 
 fiExpr platform to_drop ann_expr@(_,AnnApp {})
-  = wrapFloats drop_here $
+  = wrapFloats drop_here $ wrapFloats extra_drop $
     mkTicks ticks $
     mkApps (fiExpr platform fun_drop ann_fun)
            (zipWithEqual "fiExpr" (fiExpr platform) arg_drops ann_args)
@@ -189,18 +171,19 @@ fiExpr platform to_drop ann_expr@(_,AnnApp {})
     (ann_fun, ann_args, ticks) = collectAnnArgsTicks tickishFloatable ann_expr
     fun_ty  = exprType (deAnnotate ann_fun)
     fun_fvs = freeVarsOf ann_fun
+    arg_fvs = map freeVarsOf ann_args
 
-    (drop_here, fun_drop : arg_drops)
-       = sepBindsByDropPoint platform False to_drop
-                             here_fvs (fun_fvs : arg_fvs)
-
+    (drop_here : extra_drop : fun_drop : arg_drops)
+       = sepBindsByDropPoint platform False
+                             (extra_fvs : fun_fvs : arg_fvs)
+                             to_drop
          -- Shortcut behaviour: if to_drop is empty,
          -- sepBindsByDropPoint returns a suitable bunch of empty
          -- lists without evaluating extra_fvs, and hence without
          -- peering into each argument
 
-    ((_,here_fvs), arg_fvs) = mapAccumL add_arg (fun_ty,here_fvs0) ann_args
-    here_fvs0 = case ann_fun of
+    (_, extra_fvs) = foldl' add_arg (fun_ty, extra_fvs0) ann_args
+    extra_fvs0 = case ann_fun of
                    (_, AnnVar _) -> fun_fvs
                    _             -> emptyDVarSet
           -- Don't float the binding for f into f x y z; see Note [Join points]
@@ -208,13 +191,15 @@ fiExpr platform to_drop ann_expr@(_,AnnApp {})
           -- join point, floating it in isn't especially harmful but it's
           -- useless since the simplifier will immediately float it back out.)
 
-    add_arg :: (Type,FreeVarSet) -> CoreExprWithFVs -> ((Type,FreeVarSet),FreeVarSet)
-    add_arg (fun_ty, here_fvs) (arg_fvs, AnnType ty)
-      = ((piResultTy fun_ty ty, here_fvs), arg_fvs)
-    -- We can't float into some arguments, so put them into the here_fvs
-    add_arg (fun_ty, here_fvs) (arg_fvs, arg)
-      | noFloatIntoArg arg arg_ty = ((res_ty,here_fvs `unionDVarSet` arg_fvs), emptyDVarSet)
-      | otherwise          = ((res_ty,here_fvs), arg_fvs)
+    add_arg :: (Type,FreeVarSet) -> CoreExprWithFVs -> (Type,FreeVarSet)
+    add_arg (fun_ty, extra_fvs) (_, AnnType ty)
+      = (piResultTy fun_ty ty, extra_fvs)
+
+    add_arg (fun_ty, extra_fvs) (arg_fvs, arg)
+      | noFloatIntoArg arg arg_ty
+      = (res_ty, extra_fvs `unionDVarSet` arg_fvs)
+      | otherwise
+      = (res_ty, extra_fvs)
       where
        (_, arg_ty, res_ty) = splitFunTy fun_ty
 
@@ -298,6 +283,7 @@ it's non-recursive, so we float only into non-recursive join points.)
 Urk! if all are tyvars, and we don't float in, we may miss an
       opportunity to float inside a nested case branch
 
+
 Note [Floating coercions]
 ~~~~~~~~~~~~~~~~~~~~~~~~~
 We could, in principle, have a coercion binding like
@@ -317,36 +303,6 @@ of the types of all the drop points involved. If any of the floaters
 bind a coercion variable mentioned in any of the types, that binder must
 be dropped right away.
 
-Note [Shadowing and name capture]
-~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
-Suppose we have
-    let x = y+1 in
-    case p of
-       (y:ys) -> ...x...
-       [] -> blah
-It is obviously bogus for FloatIn to transform to
-    case p of
-       (y:ys) -> ...(let x = y+1 in x)...
-       [] -> blah
-because the y is captured.  This doesn't happen much, because shadowing is
-rare, but it did happen in #22662.
-
-One solution would be to clone as we go.  But a simpler one is this:
-
-  at a binding site (like that for (y:ys) above), abandon float-in for
-  any floating bindings that mention the binders (y, ys in this case)
-
-We achieve that by calling sepBindsByDropPoint with the binders in
-the "used-here" set:
-
-* In fiExpr (AnnLam ...).  For the body there is no need to delete
-  the lambda-binders from the body_fvs, because any bindings that
-  mention these binders will be dropped here anyway.
-
-* In fiExpr (AnnCase ...). Remember to include the case_bndr in the
-  binders.  Again, no need to delete the alt binders from the rhs
-  free vars, beause any bindings mentioning them will be dropped
-  here unconditionally.
 -}
 
 fiExpr platform to_drop lam@(_, AnnLam _ _)
@@ -355,17 +311,10 @@ fiExpr platform to_drop lam@(_, AnnLam _ _)
   = wrapFloats to_drop (mkLams bndrs (fiExpr platform [] body))
 
   | otherwise           -- Float inside
-  = wrapFloats drop_here $
-    mkLams bndrs (fiExpr platform body_drop body)
+  = mkLams bndrs (fiExpr platform to_drop body)
 
   where
     (bndrs, body) = collectAnnBndrs lam
-    body_fvs      = freeVarsOf body
-
-    -- Why sepBindsByDropPoint? Because of potential capture
-    -- See Note [Shadowing and name capture]
-    (drop_here, [body_drop]) = sepBindsByDropPoint platform False to_drop
-                                  (mkDVarSet bndrs) [body_fvs]
 
 {-
 We don't float lets inwards past an SCC.
@@ -404,7 +353,7 @@ So: rather than drop \tr{w}'s binding here, we add it onto the list of
 things to drop in the outer let's body, and let nature take its
 course.
 
-Note [extra_fvs (1): avoid floating into RHS]
+Note [extra_fvs (1)]: avoid floating into RHS
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 Consider let x=\y....t... in body.  We do not necessarily want to float
 a binding for t into the RHS, because it'll immediately be floated out
@@ -422,7 +371,7 @@ can't have unboxed bindings.
 So we make "extra_fvs" which is the rhs_fvs of such bindings, and
 arrange to dump bindings that bind extra_fvs before the entire let.
 
-Note [extra_fvs (2): free variables of rules]
+Note [extra_fvs (2)]: free variables of rules
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 Consider
   let x{rule mentioning y} = rhs in body
@@ -499,21 +448,22 @@ bindings are:
 
 fiExpr platform to_drop (_, AnnCase scrut case_bndr _ [AnnAlt con alt_bndrs rhs])
   | isUnliftedType (idType case_bndr)
+     -- binders have a fixed RuntimeRep so it's OK to call isUnliftedType
   , exprOkForSideEffects (deAnnotate scrut)
       -- See Note [Floating primops]
   = wrapFloats shared_binds $
     fiExpr platform (case_float : rhs_binds) rhs
   where
-    case_float = FB all_bndrs scrut_fvs
+    case_float = FB (mkDVarSet (case_bndr : alt_bndrs)) scrut_fvs
                     (FloatCase scrut' case_bndr con alt_bndrs)
     scrut'     = fiExpr platform scrut_binds scrut
-    rhs_fvs    = freeVarsOf rhs    -- No need to delete alt_bndrs
-    scrut_fvs  = freeVarsOf scrut  -- See Note [Shadowing and name capture]
-    all_bndrs  = mkDVarSet alt_bndrs `extendDVarSet` case_bndr
+    rhs_fvs    = freeVarsOf rhs `delDVarSetList` (case_bndr : alt_bndrs)
+    scrut_fvs  = freeVarsOf scrut
 
-    (shared_binds, [scrut_binds, rhs_binds])
-       = sepBindsByDropPoint platform False to_drop
-                     all_bndrs [scrut_fvs, rhs_fvs]
+    [shared_binds, scrut_binds, rhs_binds]
+       = sepBindsByDropPoint platform False
+           [scrut_fvs, rhs_fvs]
+           to_drop
 
 fiExpr platform to_drop (_, AnnCase scrut case_bndr ty alts)
   = wrapFloats drop_here1 $
@@ -523,43 +473,39 @@ fiExpr platform to_drop (_, AnnCase scrut case_bndr ty alts)
          -- use zipWithEqual, we should have length alts_drops_s = length alts
   where
         -- Float into the scrut and alts-considered-together just like App
-    (drop_here1, [scrut_drops, alts_drops])
-       = sepBindsByDropPoint platform False to_drop
-             all_alt_bndrs [scrut_fvs, all_alt_fvs]
-             -- all_alt_bndrs: see Note [Shadowing and name capture]
+    [drop_here1, scrut_drops, alts_drops]
+       = sepBindsByDropPoint platform False
+           [scrut_fvs, all_alts_fvs]
+           to_drop
 
         -- Float into the alts with the is_case flag set
-    (drop_here2, alts_drops_s)
-       = sepBindsByDropPoint platform True alts_drops emptyDVarSet alts_fvs
+    (drop_here2 : alts_drops_s)
+      | [ _ ] <- alts = [] : [alts_drops]
+      | otherwise     = sepBindsByDropPoint platform True alts_fvs alts_drops
 
-    scrut_fvs = freeVarsOf scrut
-
-    all_alt_bndrs = foldr (unionDVarSet . ann_alt_bndrs) (unitDVarSet case_bndr) alts
-    ann_alt_bndrs (AnnAlt _ bndrs _) = mkDVarSet bndrs
-
-    alts_fvs :: [DVarSet]
-    alts_fvs = [freeVarsOf rhs | AnnAlt _ _ rhs <- alts]
-               -- No need to delete binders
-               -- See Note [Shadowing and name capture]
-
-    all_alt_fvs :: DVarSet
-    all_alt_fvs = foldr unionDVarSet (unitDVarSet case_bndr) alts_fvs
+    scrut_fvs    = freeVarsOf scrut
+    alts_fvs     = map alt_fvs alts
+    all_alts_fvs = unionDVarSets alts_fvs
+    alt_fvs (AnnAlt _con args rhs)
+      = foldl' delDVarSet (freeVarsOf rhs) (case_bndr:args)
+           -- Delete case_bndr and args from free vars of rhs
+           -- to get free vars of alt
 
     fi_alt to_drop (AnnAlt con args rhs) = Alt con args (fiExpr platform to_drop rhs)
 
 ------------------
 fiBind :: Platform
-       -> RevFloatInBinds    -- Binds we're trying to drop
-                             -- as far "inwards" as possible
-       -> CoreBindWithFVs    -- Input binding
-       -> DVarSet            -- Free in scope of binding
-       -> ( RevFloatInBinds  -- Land these before
-          , FloatInBind      -- The binding itself
-          , RevFloatInBinds) -- Land these after
+       -> FloatInBinds      -- Binds we're trying to drop
+                            -- as far "inwards" as possible
+       -> CoreBindWithFVs   -- Input binding
+       -> DVarSet           -- Free in scope of binding
+       -> ( FloatInBinds    -- Land these before
+          , FloatInBind     -- The binding itself
+          , FloatInBinds)   -- Land these after
 
 fiBind platform to_drop (AnnNonRec id ann_rhs@(rhs_fvs, rhs)) body_fvs
-  = ( shared_binds          -- Land these before
-                                           -- See Note [extra_fvs (1,2)]
+  = ( extra_binds ++ shared_binds          -- Land these before
+                                           -- See Note [extra_fvs (1)] and Note [extra_fvs (2)]
     , FB (unitDVarSet id) rhs_fvs'         -- The new binding itself
           (FloatLet (NonRec id rhs'))
     , body_binds )                         -- Land these after
@@ -567,19 +513,20 @@ fiBind platform to_drop (AnnNonRec id ann_rhs@(rhs_fvs, rhs)) body_fvs
   where
     body_fvs2 = body_fvs `delDVarSet` id
 
-    rule_fvs = bndrRuleAndUnfoldingVarsDSet id        -- See Note [extra_fvs (2): free variables of rules]
+    rule_fvs = bndrRuleAndUnfoldingVarsDSet id        -- See Note [extra_fvs (2)]
     extra_fvs | noFloatIntoRhs NonRecursive id rhs
               = rule_fvs `unionDVarSet` rhs_fvs
               | otherwise
               = rule_fvs
-        -- See Note [extra_fvs (1): avoid floating into RHS]
+        -- See Note [extra_fvs (1)]
         -- No point in floating in only to float straight out again
         -- We *can't* float into ok-for-speculation unlifted RHSs
         -- But do float into join points
 
-    (shared_binds, [rhs_binds, body_binds])
-        = sepBindsByDropPoint platform False to_drop
-                      extra_fvs [rhs_fvs, body_fvs2]
+    [shared_binds, extra_binds, rhs_binds, body_binds]
+        = sepBindsByDropPoint platform False
+            [extra_fvs, rhs_fvs, body_fvs2]
+            to_drop
 
         -- Push rhs_binds into the right hand side of the binding
     rhs'     = fiRhs platform rhs_binds id ann_rhs
@@ -587,7 +534,7 @@ fiBind platform to_drop (AnnNonRec id ann_rhs@(rhs_fvs, rhs)) body_fvs
                         -- Don't forget the rule_fvs; the binding mentions them!
 
 fiBind platform to_drop (AnnRec bindings) body_fvs
-  = ( shared_binds
+  = ( extra_binds ++ shared_binds
     , FB (mkDVarSet ids) rhs_fvs'
          (FloatLet (Rec (fi_bind rhss_binds bindings)))
     , body_binds )
@@ -595,22 +542,23 @@ fiBind platform to_drop (AnnRec bindings) body_fvs
     (ids, rhss) = unzip bindings
     rhss_fvs = map freeVarsOf rhss
 
-        -- See Note [extra_fvs (1,2)]
+        -- See Note [extra_fvs (1)] and Note [extra_fvs (2)]
     rule_fvs = mapUnionDVarSet bndrRuleAndUnfoldingVarsDSet ids
     extra_fvs = rule_fvs `unionDVarSet`
                 unionDVarSets [ rhs_fvs | (bndr, (rhs_fvs, rhs)) <- bindings
                               , noFloatIntoRhs Recursive bndr rhs ]
 
-    (shared_binds, body_binds:rhss_binds)
-        = sepBindsByDropPoint platform False to_drop
-                       extra_fvs (body_fvs:rhss_fvs)
+    (shared_binds:extra_binds:body_binds:rhss_binds)
+        = sepBindsByDropPoint platform False
+            (extra_fvs:body_fvs:rhss_fvs)
+            to_drop
 
     rhs_fvs' = unionDVarSets rhss_fvs `unionDVarSet`
                unionDVarSets (map floatedBindsFVs rhss_binds) `unionDVarSet`
                rule_fvs         -- Don't forget the rule variables!
 
     -- Push rhs_binds into the right hand side of the binding
-    fi_bind :: [RevFloatInBinds]   -- One per "drop pt" conjured w/ fvs_of_rhss
+    fi_bind :: [FloatInBinds]       -- one per "drop pt" conjured w/ fvs_of_rhss
             -> [(Id, CoreExprWithFVs)]
             -> [(Id, CoreExpr)]
 
@@ -619,7 +567,7 @@ fiBind platform to_drop (AnnRec bindings) body_fvs
         | ((binder, rhs), to_drop) <- zipEqual "fi_bind" pairs to_drops ]
 
 ------------------
-fiRhs :: Platform -> RevFloatInBinds -> CoreBndr -> CoreExprWithFVs -> CoreExpr
+fiRhs :: Platform -> FloatInBinds -> CoreBndr -> CoreExprWithFVs -> CoreExpr
 fiRhs platform to_drop bndr rhs
   | Just join_arity <- isJoinId_maybe bndr
   , let (bndrs, body) = collectNAnnBndrs join_arity rhs
@@ -645,7 +593,7 @@ noFloatIntoRhs is_rec bndr rhs
 
 noFloatIntoArg :: CoreExprWithFVs' -> Type -> Bool
 noFloatIntoArg expr expr_ty
-  | isUnliftedType expr_ty
+  | Just Unlifted <- typeLevity_maybe expr_ty
   = True  -- See Note [Do not destroy the let/app invariant]
 
    | AnnLam bndr e <- expr
@@ -654,7 +602,7 @@ noFloatIntoArg expr expr_ty
    || all isTyVar (bndr:bndrs)     -- Wrinkle 1 (b)
       -- See Note [noFloatInto considerations] wrinkle 2
 
-  | otherwise  -- Note [noFloatInto considerations] wrinkle 2
+  | otherwise  -- See Note [noFloatInto considerations] wrinkle 2
   = exprIsTrivial deann_expr || exprIsHNF deann_expr
   where
     deann_expr = deAnnotate' expr
@@ -719,84 +667,68 @@ point.
 We have to maintain the order on these drop-point-related lists.
 -}
 
--- pprFIB :: RevFloatInBinds -> SDoc
+-- pprFIB :: FloatInBinds -> SDoc
 -- pprFIB fibs = text "FIB:" <+> ppr [b | FB _ _ b <- fibs]
 
 sepBindsByDropPoint
     :: Platform
-    -> Bool                  -- True <=> is case expression
-    -> RevFloatInBinds       -- Candidate floaters
-    -> FreeVarSet            -- here_fvs: if these vars are free in a binding,
-                             --   don't float that binding inside any drop point
-    -> [FreeVarSet]          -- fork_fvs: one set of FVs per drop point
-    -> ( RevFloatInBinds     -- Bindings which must not be floated inside
-       , [RevFloatInBinds] ) -- Corresponds 1-1 with the input list of FV sets
+    -> Bool                -- True <=> is case expression
+    -> [FreeVarSet]        -- One set of FVs per drop point
+                           -- Always at least two long!
+    -> FloatInBinds        -- Candidate floaters
+    -> [FloatInBinds]      -- FIRST one is bindings which must not be floated
+                           -- inside any drop point; the rest correspond
+                           -- one-to-one with the input list of FV sets
 
 -- Every input floater is returned somewhere in the result;
 -- none are dropped, not even ones which don't seem to be
 -- free in *any* of the drop-point fvs.  Why?  Because, for example,
 -- a binding (let x = E in B) might have a specialised version of
 -- x (say x') stored inside x, but x' isn't free in E or B.
---
--- The here_fvs argument is used for two things:
--- * Avoid shadowing bugs: see Note [Shadowing and name capture]
--- * Drop some of the bindings at the top, e.g. of an application
 
 type DropBox = (FreeVarSet, FloatInBinds)
 
-dropBoxFloats :: DropBox -> RevFloatInBinds
-dropBoxFloats (_, floats) = reverse floats
-
-usedInDropBox :: DIdSet -> DropBox -> Bool
-usedInDropBox bndrs (db_fvs, _) = db_fvs `intersectsDVarSet` bndrs
-
-initDropBox :: DVarSet -> DropBox
-initDropBox fvs = (fvs, [])
-
-sepBindsByDropPoint platform is_case floaters here_fvs fork_fvs
+sepBindsByDropPoint platform is_case drop_pts floaters
   | null floaters  -- Shortcut common case
-  = ([], [[] | _ <- fork_fvs])
+  = [] : [[] | _ <- drop_pts]
 
   | otherwise
-  = go floaters (initDropBox here_fvs) (map initDropBox fork_fvs)
+  = assert (drop_pts `lengthAtLeast` 2) $
+    go floaters (map (\fvs -> (fvs, [])) (emptyDVarSet : drop_pts))
   where
-    n_alts = length fork_fvs
+    n_alts = length drop_pts
 
-    go :: RevFloatInBinds -> DropBox -> [DropBox]
-       -> (RevFloatInBinds, [RevFloatInBinds])
-        -- The *first* one in the pair is the drop_here set
+    go :: FloatInBinds -> [DropBox] -> [FloatInBinds]
+        -- The *first* one in the argument list is the drop_here set
+        -- The FloatInBinds in the lists are in the reverse of
+        -- the normal FloatInBinds order; that is, they are the right way round!
 
-    go [] here_box fork_boxes
-        = (dropBoxFloats here_box, map dropBoxFloats fork_boxes)
+    go [] drop_boxes = map (reverse . snd) drop_boxes
 
-    go (bind_w_fvs@(FB bndrs bind_fvs bind) : binds) here_box fork_boxes
-        | drop_here = go binds (insert here_box) fork_boxes
-        | otherwise = go binds here_box          new_fork_boxes
+    go (bind_w_fvs@(FB bndrs bind_fvs bind) : binds) drop_boxes@(here_box : fork_boxes)
+        = go binds new_boxes
         where
           -- "here" means the group of bindings dropped at the top of the fork
 
-          used_here     = bndrs `usedInDropBox` here_box
-          used_in_flags = case fork_boxes of
-                            []  -> []
-                            [_] -> [True]  -- Push all bindings into a single branch
-                                           -- No need to look at its free vars
-                            _   -> map (bndrs `usedInDropBox`) fork_boxes
-               -- Short-cut for the singleton case;
-               -- used for lambdas and singleton cases
+          (used_here : used_in_flags) = [ fvs `intersectsDVarSet` bndrs
+                                        | (fvs, _) <- drop_boxes]
 
           drop_here = used_here || cant_push
 
           n_used_alts = count id used_in_flags -- returns number of Trues in list.
 
           cant_push
-            | is_case   = (n_alts > 1 && n_used_alts == n_alts)
-                             -- Used in all, muliple branches, don't push
+            | is_case   = n_used_alts == n_alts   -- Used in all, don't push
+                                                  -- Remember n_alts > 1
                           || (n_used_alts > 1 && not (floatIsDupable platform bind))
                              -- floatIsDupable: see Note [Duplicating floats]
 
             | otherwise = floatIsCase bind || n_used_alts > 1
                              -- floatIsCase: see Note [Floating primops]
 
+          new_boxes | drop_here = (insert here_box : fork_boxes)
+                    | otherwise = (here_box : new_fork_boxes)
+
           new_fork_boxes = zipWithEqual "FloatIn.sepBinds" insert_maybe
                                         fork_boxes used_in_flags
 
@@ -806,10 +738,11 @@ sepBindsByDropPoint platform is_case floaters here_fvs fork_fvs
           insert_maybe box True  = insert box
           insert_maybe box False = box
 
+    go _ _ = panic "sepBindsByDropPoint/go"
+
 
 {- Note [Duplicating floats]
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
-
 For case expressions we duplicate the binding if it is reasonably
 small, and if it is not used in all the RHSs This is good for
 situations like
@@ -823,14 +756,14 @@ If the thing is used in all RHSs there is nothing gained,
 so we don't duplicate then.
 -}
 
-floatedBindsFVs :: RevFloatInBinds -> FreeVarSet
+floatedBindsFVs :: FloatInBinds -> FreeVarSet
 floatedBindsFVs binds = mapUnionDVarSet fbFVs binds
 
 fbFVs :: FloatInBind -> DVarSet
 fbFVs (FB _ fvs _) = fvs
 
-wrapFloats :: RevFloatInBinds -> CoreExpr -> CoreExpr
--- Remember RevFloatInBinds is in *reverse* dependency order
+wrapFloats :: FloatInBinds -> CoreExpr -> CoreExpr
+-- Remember FloatInBinds is in *reverse* dependency order
 wrapFloats []               e = e
 wrapFloats (FB _ _ fl : bs) e = wrapFloats bs (wrapFloat fl e)
 
