@@ -1,8 +1,8 @@
 -- (c) The University of Glasgow 2006
 
 {-# LANGUAGE ScopedTypeVariables, PatternSynonyms #-}
-{-# LANGUAGE CPP #-}
-{-# LANGUAGE DeriveFunctor, DeriveDataTypeable #-}
+
+{-# LANGUAGE DeriveFunctor #-}
 
 module GHC.Core.Unify (
         tcMatchTy, tcMatchTyKi,
@@ -11,8 +11,8 @@ module GHC.Core.Unify (
         tcMatchTyX_BM, ruleMatchTyKiX,
 
         -- * Rough matching
-        RoughMatchTc(..), roughMatchTcs, instanceCantMatch,
-        typesCantMatch, isRoughOtherTc,
+        RoughMatchTc(..), roughMatchTcs, roughMatchTcsLookup, instanceCantMatch,
+        typesCantMatch, isRoughWildcard,
 
         -- Side-effect free unification
         tcUnifyTy, tcUnifyTyKi, tcUnifyTys, tcUnifyTyKis,
@@ -27,8 +27,6 @@ module GHC.Core.Unify (
         flattenTys, flattenTysX
    ) where
 
-#include "HsVersions.h"
-
 import GHC.Prelude
 
 import GHC.Types.Var
@@ -41,6 +39,7 @@ import GHC.Core.TyCon
 import GHC.Core.TyCo.Rep
 import GHC.Core.TyCo.FVs   ( tyCoVarsOfCoList, tyCoFVsOfTypes )
 import GHC.Core.TyCo.Subst ( mkTvSubst )
+import GHC.Core.RoughMap
 import GHC.Core.Map.Type
 import GHC.Utils.FV( FV, fvVarSet, fvVarList )
 import GHC.Utils.Misc
@@ -49,15 +48,17 @@ import GHC.Utils.Outputable
 import GHC.Types.Unique
 import GHC.Types.Unique.FM
 import GHC.Types.Unique.Set
+import {-# SOURCE #-} GHC.Tc.Utils.TcType ( tcEqType )
 import GHC.Exts( oneShot )
-import GHC.Utils.Panic
+import GHC.Utils.Panic.Plain
 import GHC.Data.FastString
 
-import Data.Data ( Data )
 import Data.List ( mapAccumL )
 import Control.Monad
 import qualified Data.Semigroup as S
 
+import GHC.Builtin.Names (constraintKindTyConKey, liftedTypeKindTyConKey)
+
 {-
 
 Unification is much tricker than you might think.
@@ -291,27 +292,11 @@ But it is never
              albeit perhaps only after 'a' is instantiated.
 -}
 
-data RoughMatchTc
-  = KnownTc Name   -- INVARIANT: Name refers to a TyCon tc that responds
-                   -- true to `isGenerativeTyCon tc Nominal`. See
-                   -- Note [Rough matching in class and family instances]
-  | OtherTc        -- e.g. type variable at the head
-  deriving( Data )
-
-isRoughOtherTc :: RoughMatchTc -> Bool
-isRoughOtherTc OtherTc      = True
-isRoughOtherTc (KnownTc {}) = False
-
 roughMatchTcs :: [Type] -> [RoughMatchTc]
-roughMatchTcs tys = map rough tys
-  where
-    rough ty
-      | Just (ty', _) <- splitCastTy_maybe ty   = rough ty'
-      | Just (tc,_)   <- splitTyConApp_maybe ty
-      , not (isTypeFamilyTyCon tc)              = ASSERT2( isGenerativeTyCon tc Nominal, ppr tc )
-                                                  KnownTc (tyConName tc)
-        -- See Note [Rough matching in class and family instances]
-      | otherwise                               = OtherTc
+roughMatchTcs tys = map typeToRoughMatchTc tys
+
+roughMatchTcsLookup :: [Type] -> [RoughMatchLookupTc]
+roughMatchTcsLookup tys = map typeToRoughMatchLookupTc tys
 
 instanceCantMatch :: [RoughMatchTc] -> [RoughMatchTc] -> Bool
 -- (instanceCantMatch tcs1 tcs2) returns True if tcs1 cannot
@@ -321,7 +306,7 @@ instanceCantMatch (mt : ts) (ma : as) = itemCantMatch mt ma || instanceCantMatch
 instanceCantMatch _         _         =  False  -- Safe
 
 itemCantMatch :: RoughMatchTc -> RoughMatchTc -> Bool
-itemCantMatch (KnownTc t) (KnownTc a) = t /= a
+itemCantMatch (RM_KnownTc t) (RM_KnownTc a) = t /= a
 itemCantMatch _           _           = False
 
 
@@ -414,8 +399,8 @@ complete. This means that, sometimes, a closed type family does not reduce
 when it should. See test case indexed-types/should_fail/Overlap15 for an
 example.
 
-Note [Unificiation result]
-~~~~~~~~~~~~~~~~~~~~~~~~~~
+Note [Unification result]
+~~~~~~~~~~~~~~~~~~~~~~~~~
 When unifying t1 ~ t2, we return
 * Unifiable s, if s is a substitution such that s(t1) is syntactically the
   same as s(t2), modulo type-synonym expansion.
@@ -539,7 +524,7 @@ tcUnifyTyKis bind_fn tys1 tys2
 -- return the final result. See Note [Fine-grained unification]
 type UnifyResult = UnifyResultM TCvSubst
 
--- | See Note [Unificiation result]
+-- | See Note [Unification result]
 data UnifyResultM a = Unifiable a        -- the subst that unifies the types
                     | MaybeApart MaybeApartReason
                                  a       -- the subst has as much as we know
@@ -548,19 +533,25 @@ data UnifyResultM a = Unifiable a        -- the subst that unifies the types
                     | SurelyApart
                     deriving Functor
 
--- | Why are two types 'MaybeApart'? 'MARTypeFamily' takes precedence:
+-- | Why are two types 'MaybeApart'? 'MARInfinite' takes precedence:
 -- This is used (only) in Note [Infinitary substitution in lookup] in GHC.Core.InstEnv
+-- As of Feb 2022, we never differentiate between MARTypeFamily and MARTypeVsConstraint;
+-- it's really only MARInfinite that's interesting here.
 data MaybeApartReason = MARTypeFamily   -- ^ matching e.g. F Int ~? Bool
                       | MARInfinite     -- ^ matching e.g. a ~? Maybe a
+                      | MARTypeVsConstraint  -- ^ matching Type ~? Constraint
+                                             -- See Note [coreView vs tcView] in GHC.Core.Type
 
 instance Outputable MaybeApartReason where
-  ppr MARTypeFamily = text "MARTypeFamily"
-  ppr MARInfinite   = text "MARInfinite"
+  ppr MARTypeFamily       = text "MARTypeFamily"
+  ppr MARInfinite         = text "MARInfinite"
+  ppr MARTypeVsConstraint = text "MARTypeVsConstraint"
 
 instance Semigroup MaybeApartReason where
   -- see end of Note [Unification result] for why
-  MARTypeFamily <> r = r
-  MARInfinite   <> _ = MARInfinite
+  MARTypeFamily       <> r = r
+  MARInfinite         <> _ = MARInfinite
+  MARTypeVsConstraint <> r = r
 
 instance Applicative UnifyResultM where
   pure  = Unifiable
@@ -1069,13 +1060,22 @@ unify_ty :: UMEnv
 -- See Note [Specification of unification]
 -- Respects newtypes, PredTypes
 -- See Note [Computing equality on types] in GHC.Core.Type
-unify_ty env ty1 ty2 kco
+unify_ty _env (TyConApp tc1 []) (TyConApp tc2 []) _kco
   -- See Note [Comparing nullary type synonyms] in GHC.Core.Type.
-  | TyConApp tc1 [] <- ty1
-  , TyConApp tc2 [] <- ty2
-  , tc1 == tc2                = return ()
+  | tc1 == tc2
+  = return ()
 
-    -- TODO: More commentary needed here
+  -- See Note [coreView vs tcView] in GHC.Core.Type.
+  | tc1 `hasKey` constraintKindTyConKey
+  , tc2 `hasKey` liftedTypeKindTyConKey
+  = maybeApart MARTypeVsConstraint
+
+  | tc2 `hasKey` constraintKindTyConKey
+  , tc1 `hasKey` liftedTypeKindTyConKey
+  = maybeApart MARTypeVsConstraint
+
+unify_ty env ty1 ty2 kco
+    -- Now handle the cases we can "look through": synonyms and casts.
   | Just ty1' <- tcView ty1   = unify_ty env ty1' ty2 kco
   | Just ty2' <- tcView ty2   = unify_ty env ty1 ty2' kco
   | CastTy ty1' co <- ty1     = if um_unif env
@@ -1114,11 +1114,20 @@ unify_ty env ty1 ty2 _kco
             ; unless (um_inj_tf env) $ -- See (end of) Note [Specification of unification]
               don'tBeSoSure MARTypeFamily $ unify_tys env noninj_tys1 noninj_tys2 }
 
-  | isTyFamApp mb_tc_app1     -- A (not-over-saturated) type-family application
-  = maybeApart MARTypeFamily  -- behaves like a type variable; might match
+  | Just (tc1, _) <- mb_tc_app1
+  , not (isGenerativeTyCon tc1 Nominal)
+    -- E.g.   unify_ty (F ty1) b  =  MaybeApart
+    --        because the (F ty1) behaves like a variable
+    --        NB: if unifying, we have already dealt
+    --            with the 'ty2 = variable' case
+  = maybeApart MARTypeFamily
 
-  | isTyFamApp mb_tc_app2     -- A (not-over-saturated) type-family application
-  , um_unif env               -- behaves like a type variable; might unify
+  | Just (tc2, _) <- mb_tc_app2
+  , not (isGenerativeTyCon tc2 Nominal)
+  , um_unif env
+    -- E.g.   unify_ty [a] (F ty2) =  MaybeApart, when unifying (only)
+    --        because the (F ty2) behaves like a variable
+    --        NB: we have already dealt with the 'ty1 = variable' case
   = maybeApart MARTypeFamily
 
   where
@@ -1207,17 +1216,6 @@ unify_tys env orig_xs orig_ys
       -- Possibly different saturations of a polykinded tycon
       -- See Note [Polykinded tycon applications]
 
-isTyFamApp :: Maybe (TyCon, [Type]) -> Bool
--- True if we have a saturated or under-saturated type family application
--- If it is /over/ saturated then we return False.  E.g.
---     unify_ty (F a b) (c d)    where F has arity 1
--- we definitely want to decompose that type application! (#22647)
-isTyFamApp (Just (tc, tys))
-  =  not (isGenerativeTyCon tc Nominal)       -- Type family-ish
-  && not (tys `lengthExceeds` tyConArity tc)  -- Not over-saturated
-isTyFamApp Nothing
-  = False
-
 ---------------------------------
 uVar :: UMEnv
      -> InTyVar         -- Variable to be unified
@@ -1239,8 +1237,17 @@ uVar env tv1 ty kco
                       -- this is because the range of the subst is the target
                       -- type, not the template type. So, just check for
                       -- normal type equality.
-                      unless ((ty' `mkCastTy` kco) `eqType` ty) $
-                      surelyApart
+                      unless ((ty' `mkCastTy` kco) `tcEqType` ty) $
+                        surelyApart
+                      -- NB: it's important to use `tcEqType` instead of `eqType` here,
+                      -- otherwise we might not reject a substitution
+                      -- which unifies `Type` with `Constraint`, e.g.
+                      -- a call to tc_unify_tys with arguments
+                      --
+                      --   tys1 = [k,k]
+                      --   tys2 = [Type, Constraint]
+                      --
+                      -- See test cases: T11715b, T20521.
           Nothing  -> uUnrefined env tv1' ty ty kco } -- No, continue
 
 uUnrefined :: UMEnv
@@ -1396,6 +1403,8 @@ instance Applicative UM where
       (<*>)  = ap
 
 instance Monad UM where
+  {-# INLINE (>>=) #-}
+  -- See Note [INLINE pragmas and (>>)] in GHC.Utils.Monad
   m >>= k  = UM (\state ->
                   do { (state', v) <- unUM m state
                      ; unUM (k v) state' })
@@ -1537,10 +1546,13 @@ liftCoMatch tmpls ty co
 ty_co_match :: MatchEnv   -- ^ ambient helpful info
             -> LiftCoEnv  -- ^ incoming subst
             -> Type       -- ^ ty, type to match
-            -> Coercion   -- ^ co, coercion to match against
-            -> Coercion   -- ^ :: kind of L type of substed ty ~N L kind of co
-            -> Coercion   -- ^ :: kind of R type of substed ty ~N R kind of co
+            -> Coercion   -- ^ co :: lty ~r rty, coercion to match against
+            -> Coercion   -- ^ :: kind(lsubst(ty)) ~N kind(lty)
+            -> Coercion   -- ^ :: kind(rsubst(ty)) ~N kind(rty)
             -> Maybe LiftCoEnv
+   -- ^ Just env ==> liftCoSubst Nominal env ty == co, modulo roles.
+   -- Also: Just env ==> lsubst(ty) == lty and rsubst(ty) == rty,
+   -- where lsubst = lcSubstLeft(env) and rsubst = lcSubstRight(env)
 ty_co_match menv subst ty co lkco rkco
   | Just ty' <- coreView ty = ty_co_match menv subst ty' co lkco rkco
      -- why coreView here, not tcView? Because we're firmly after type-checking.
@@ -1550,6 +1562,8 @@ ty_co_match menv subst ty co lkco rkco
   | tyCoVarsOfType ty `isNotInDomainOf` subst
   , Just (ty', _) <- isReflCo_maybe co
   , ty `eqType` ty'
+    -- Why `eqType` and not `tcEqType`? Because this function is only used
+    -- during coercion optimisation, after type-checking has finished.
   = Just subst
 
   where
@@ -1608,14 +1622,17 @@ ty_co_match menv subst ty1 (AppCo co2 arg2) _lkco _rkco
 ty_co_match menv subst (TyConApp tc1 tys) (TyConAppCo _ tc2 cos) _lkco _rkco
   = ty_co_match_tc menv subst tc1 tys tc2 cos
 ty_co_match menv subst (FunTy _ w ty1 ty2) co _lkco _rkco
-    -- Despite the fact that (->) is polymorphic in five type variables (two
-    -- runtime rep, a multiplicity and two types), we shouldn't need to
-    -- explicitly unify the runtime reps here; unifying the types themselves
-    -- should be sufficient.  See Note [Representation of function types].
-  | Just (tc, [co_mult, _,_,co1,co2]) <- splitTyConAppCo_maybe co
+  | Just (tc, [co_mult,rrco1,rrco2,co1,co2]) <- splitTyConAppCo_maybe co
   , tc == funTyCon
-  = let Pair lkcos rkcos = traverse (fmap mkNomReflCo . coercionKind) [co_mult,co1,co2]
-    in ty_co_match_args menv subst [w, ty1, ty2] [co_mult, co1, co2] lkcos rkcos
+  = let rr1 = getRuntimeRep ty1
+        rr2 = getRuntimeRep ty2
+        Pair lkcos rkcos = traverse (fmap (mkNomReflCo . typeKind) . coercionKind)
+                             [co_mult,rrco1, rrco2,co1,co2]
+    in  -- NB: we include the RuntimeRep arguments in the matching; not doing so caused #21205.
+        ty_co_match_args menv subst
+          [w, rr1, rr2, ty1, ty2]
+          [co_mult, rrco1, rrco2, co1, co2]
+          lkcos rkcos
 
 ty_co_match menv subst (ForAllTy (Bndr tv1 _) ty1)
                        (ForAllCo tv2 kind_co2 co2)
@@ -1684,7 +1701,7 @@ ty_co_match_tc menv subst tc1 tys1 tc2 cos2
        ; ty_co_match_args menv subst tys1 cos2 lkcos rkcos }
   where
     Pair lkcos rkcos
-      = traverse (fmap mkNomReflCo . coercionKind) cos2
+      = traverse (fmap (mkNomReflCo . typeKind) . coercionKind) cos2
 
 ty_co_match_app :: MatchEnv -> LiftCoEnv
                 -> Type -> [Type] -> Coercion -> [Coercion]
@@ -1698,7 +1715,7 @@ ty_co_match_app menv subst ty1 ty1args co2 co2args
   = do { subst1 <- ty_co_match menv subst ki1 ki2 ki_ki_co ki_ki_co
        ; let Pair lkco rkco = mkNomReflCo <$> coercionKind ki2
        ; subst2 <- ty_co_match menv subst1 ty1 co2 lkco rkco
-       ; let Pair lkcos rkcos = traverse (fmap mkNomReflCo . coercionKind) co2args
+       ; let Pair lkcos rkcos = traverse (fmap (mkNomReflCo . typeKind) . coercionKind) co2args
        ; ty_co_match_args menv subst2 ty1args co2args lkcos rkcos }
   where
     ki1 = typeKind ty1
@@ -2023,7 +2040,7 @@ coreFlattenTyFamApp tv_subst env fam_tc fam_args
   where
     arity = tyConArity fam_tc
     tcv_subst = TCvSubst (fe_in_scope env) tv_subst emptyVarEnv
-    (sat_fam_args, leftover_args) = ASSERT( arity <= length fam_args )
+    (sat_fam_args, leftover_args) = assert (arity <= length fam_args) $
                                     splitAt arity fam_args
     -- Apply the substitution before looking up an application in the
     -- environment. See Note [Flattening type-family applications when matching instances],
