@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP #-}
+
 {-# LANGUAGE Strict #-} -- See Note [Avoiding space leaks in toIface*]
 
 -- | Functions for converting Core things to interface file things.
@@ -44,42 +44,45 @@ module GHC.CoreToIface
     , toIfaceLFInfo
     ) where
 
-#include "HsVersions.h"
-
 import GHC.Prelude
 
-import GHC.Driver.Ppr
-import GHC.Iface.Syntax
-import GHC.Core.DataCon
-import GHC.Types.Id
-import GHC.Types.Literal
-import GHC.Types.Id.Info
 import GHC.StgToCmm.Types
+
 import GHC.Core
 import GHC.Core.TyCon hiding ( pprPromotionQuote )
 import GHC.Core.Coercion.Axiom
-import GHC.Builtin.Types.Prim ( eqPrimTyCon, eqReprPrimTyCon )
-import GHC.Builtin.Types ( heqTyCon )
-import GHC.Types.Id.Make ( noinlineIdName )
-import GHC.Builtin.Names
-import GHC.Types.Name
-import GHC.Types.Basic
+import GHC.Core.DataCon
 import GHC.Core.Type
 import GHC.Core.Multiplicity
 import GHC.Core.PatSyn
-import GHC.Utils.Outputable
-import GHC.Utils.Panic
+import GHC.Core.TyCo.Rep
+import GHC.Core.TyCo.Tidy ( tidyCo )
+
+import GHC.Builtin.Types.Prim ( eqPrimTyCon, eqReprPrimTyCon )
+import GHC.Builtin.Types ( heqTyCon )
+import GHC.Builtin.Names
+
+import GHC.Iface.Syntax
 import GHC.Data.FastString
-import GHC.Utils.Misc
+
+import GHC.Types.Id
+import GHC.Types.Id.Info
+import GHC.Types.Id.Make ( noinlineIdName )
+import GHC.Types.Literal
+import GHC.Types.Name
+import GHC.Types.Basic
 import GHC.Types.Var
 import GHC.Types.Var.Env
 import GHC.Types.Var.Set
 import GHC.Types.Tickish
-import GHC.Core.TyCo.Rep
-import GHC.Core.TyCo.Tidy ( tidyCo )
 import GHC.Types.Demand ( isTopSig )
 import GHC.Types.Cpr ( topCprSig )
 
+import GHC.Utils.Outputable
+import GHC.Utils.Panic
+import GHC.Utils.Misc
+import GHC.Utils.Trace
+
 import Data.Maybe ( catMaybes )
 
 {- Note [Avoiding space leaks in toIface*]
@@ -166,7 +169,7 @@ toIfaceTypeX :: VarSet -> Type -> IfaceType
 --    translates the tyvars in 'free' as IfaceFreeTyVars
 --
 -- Synonyms are retained in the interface type
-toIfaceTypeX fr (TyVarTy tv)   -- See Note [TcTyVars in IfaceType] in GHC.Iface.Type
+toIfaceTypeX fr (TyVarTy tv)   -- See Note [Free tyvars in IfaceType] in GHC.Iface.Type
   | tv `elemVarSet` fr         = IfaceFreeTyVar tv
   | otherwise                  = IfaceTyVar (toIfaceTyVar tv)
 toIfaceTypeX fr ty@(AppTy {})  =
@@ -279,7 +282,7 @@ toIfaceCoercionX fr co
     go (Refl ty)            = IfaceReflCo (toIfaceTypeX fr ty)
     go (GRefl r ty mco)     = IfaceGReflCo r (toIfaceTypeX fr ty) (go_mco mco)
     go (CoVarCo cv)
-      -- See [TcTyVars in IfaceType] in GHC.Iface.Type
+      -- See Note [Free tyvars in IfaceType] in GHC.Iface.Type
       | cv `elemVarSet` fr  = IfaceFreeCoVar cv
       | otherwise           = IfaceCoVarCo (toIfaceCoVar cv)
     go (HoleCo h)           = IfaceHoleCo  (coHoleCoVar h)
@@ -299,7 +302,7 @@ toIfaceCoercionX fr co
                                           (toIfaceTypeX fr t2)
     go (TyConAppCo r tc cos)
       | tc `hasKey` funTyConKey
-      , [_,_,_,_, _] <- cos         = pprPanic "toIfaceCoercion" empty
+      , [_,_,_,_, _] <- cos         = panic "toIfaceCoercion"
       | otherwise                =
         IfaceTyConAppCo r (toIfaceTyCon tc) (map go cos)
     go (FunCo r w co1 co2)   = IfaceFunCo r (go w) (go co1) (go co2)
@@ -369,7 +372,7 @@ toIfaceAppArgsX fr kind ty_args
         -- This is probably a compiler bug, so we print a trace and
         -- carry on as if it were FunTy.  Without the test for
         -- isEmptyTCvSubst we'd get an infinite loop (#15473)
-        WARN( True, ppr kind $$ ppr ty_args )
+        warnPprTrace True "toIfaceAppArgsX" (ppr kind $$ ppr ty_args) $
         IA_Arg (toIfaceTypeX fr t1) Required (go env ty ts1)
 
 tidyToIfaceType :: TidyEnv -> Type -> IfaceType
@@ -438,6 +441,7 @@ toIfaceLetBndr id  = IfLetBndr (occNameFS (getOccName id))
 
 toIfaceIdDetails :: IdDetails -> IfaceIdDetails
 toIfaceIdDetails VanillaId                      = IfVanillaId
+toIfaceIdDetails (WorkerLikeId dmds)          = IfWorkerLikeId dmds
 toIfaceIdDetails (DFunId {})                    = IfDFunId
 toIfaceIdDetails (RecSelId { sel_naughty = n
                            , sel_tycon = tc })  =
@@ -471,16 +475,16 @@ toIfaceIdInfo id_info
 
     ------------  Strictness  --------------
         -- No point in explicitly exporting TopSig
-    sig_info = strictnessInfo id_info
-    strict_hsinfo | not (isTopSig sig_info) = Just (HsStrictness sig_info)
+    sig_info = dmdSigInfo id_info
+    strict_hsinfo | not (isTopSig sig_info) = Just (HsDmdSig sig_info)
                   | otherwise               = Nothing
 
     ------------  CPR --------------
-    cpr_info = cprInfo id_info
-    cpr_hsinfo | cpr_info /= topCprSig = Just (HsCpr cpr_info)
+    cpr_info = cprSigInfo id_info
+    cpr_hsinfo | cpr_info /= topCprSig = Just (HsCprSig cpr_info)
                | otherwise             = Nothing
     ------------  Unfolding  --------------
-    unfold_hsinfo = toIfUnfolding loop_breaker (unfoldingInfo id_info)
+    unfold_hsinfo = toIfUnfolding loop_breaker (realUnfoldingInfo id_info)
     loop_breaker  = isStrongLoopBreaker (occInfo id_info)
 
     ------------  Inline prag  --------------
@@ -488,8 +492,8 @@ toIfaceIdInfo id_info
     inline_hsinfo | isDefaultInlinePragma inline_prag = Nothing
                   | otherwise = Just (HsInline inline_prag)
 
-    ------------  Levity polymorphism  ----------
-    levity_hsinfo | isNeverLevPolyIdInfo id_info = Just HsLevity
+    ------------  Representation polymorphism  ----------
+    levity_hsinfo | isNeverRepPolyIdInfo id_info = Just HsLevity
                   | otherwise                    = Nothing
 
 toIfaceJoinInfo :: Maybe JoinArity -> IfaceJoinInfo
@@ -524,7 +528,7 @@ toIfUnfolding lb (DFunUnfolding { df_bndrs = bndrs, df_args = args })
 
 toIfUnfolding _ (OtherCon {}) = Nothing
   -- The binding site of an Id doesn't have OtherCon, except perhaps
-  -- where we have called zapUnfolding; and that evald'ness info is
+  -- where we have called trimUnfolding; and that evald'ness info is
   -- not needed by importing modules
 
 toIfUnfolding _ BootUnfolding = Nothing
@@ -585,9 +589,9 @@ toIfaceAlt (Alt c bs r) = IfaceAlt (toIfaceCon c) (map getOccFS bs) (toIfaceExpr
 ---------------------
 toIfaceCon :: AltCon -> IfaceConAlt
 toIfaceCon (DataAlt dc) = IfaceDataAlt (getName dc)
-toIfaceCon (LitAlt l)   = ASSERT2( (not (isLitRubbish l)) , ppr l )
+toIfaceCon (LitAlt l)   = assertPpr (not (isLitRubbish l)) (ppr l) $
                           -- assert: see Note [Rubbish literals] wrinkle (b)
-                          (IfaceLitAlt l)
+                          IfaceLitAlt l
 toIfaceCon DEFAULT      = IfaceDefault
 
 ---------------------
@@ -635,15 +639,15 @@ toIfaceLFInfo nm lfi = case lfi of
     LFReEntrant top_lvl arity no_fvs _arg_descr ->
       -- Exported LFReEntrant closures are top level, and top-level closures
       -- don't have free variables
-      ASSERT2(isTopLevel top_lvl, ppr nm)
-      ASSERT2(no_fvs, ppr nm)
+      assertPpr (isTopLevel top_lvl) (ppr nm) $
+      assertPpr no_fvs (ppr nm) $
       IfLFReEntrant arity
     LFThunk top_lvl no_fvs updatable sfi mb_fun ->
       -- Exported LFThunk closures are top level (which don't have free
       -- variables) and non-standard (see cgTopRhsClosure)
-      ASSERT2(isTopLevel top_lvl, ppr nm)
-      ASSERT2(no_fvs, ppr nm)
-      ASSERT2(sfi == NonStandardThunk, ppr nm)
+      assertPpr (isTopLevel top_lvl) (ppr nm) $
+      assertPpr no_fvs (ppr nm) $
+      assertPpr (sfi == NonStandardThunk) (ppr nm) $
       IfLFThunk updatable mb_fun
     LFCon dc ->
       IfLFCon (dataConName dc)
