@@ -1,5 +1,5 @@
-{-# LANGUAGE BangPatterns #-}
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE BangPatterns #-}
 {-# LANGUAGE RecordWildCards #-}
 
 {-# OPTIONS_GHC -Wno-incomplete-uni-patterns #-}
@@ -44,15 +44,15 @@ module GHC.Runtime.Eval (
         Term(..), obtainTermFromId, obtainTermFromVal, reconstructType
         ) where
 
-#include "HsVersions.h"
-
 import GHC.Prelude
 
 import GHC.Driver.Monad
 import GHC.Driver.Main
+import GHC.Driver.Errors.Types ( hoistTcRnMessage )
 import GHC.Driver.Env
 import GHC.Driver.Session
 import GHC.Driver.Ppr
+import GHC.Driver.Config
 
 import GHC.Runtime.Eval.Types
 import GHC.Runtime.Interpreter as GHCi
@@ -81,7 +81,6 @@ import GHC.Tc.Types.Constraint
 import GHC.Tc.Types.Origin
 
 import GHC.Builtin.Names ( toDynName, pretendNameIsInScope )
-import GHC.Builtin.Types ( isCTupleTyConName )
 
 import GHC.Data.Maybe
 import GHC.Data.FastString
@@ -93,6 +92,7 @@ import GHC.Utils.Error
 import GHC.Utils.Outputable
 import GHC.Utils.Misc
 import GHC.Utils.Logger
+import GHC.Utils.Trace
 
 import GHC.Types.RepType
 import GHC.Types.Fixity.Env
@@ -105,7 +105,10 @@ import GHC.Types.Var.Env
 import GHC.Types.SrcLoc
 import GHC.Types.Unique
 import GHC.Types.Unique.Supply
+import GHC.Types.Unique.DSet
 import GHC.Types.TyThing
+import GHC.Types.BreakInfo
+import GHC.Types.Unique.Map
 
 import GHC.Unit
 import GHC.Unit.Module.Graph
@@ -119,7 +122,6 @@ import Data.Either
 import Data.IntMap (IntMap)
 import qualified Data.IntMap as IntMap
 import Data.List (find,intercalate)
-import qualified Data.Map as Map
 import Control.Monad
 import Control.Monad.Catch as MC
 import Data.Array
@@ -133,6 +135,7 @@ import GHC.Tc.Utils.Instantiate (instDFunType)
 import GHC.Tc.Solver (simplifyWantedsTcM)
 import GHC.Tc.Utils.Monad
 import GHC.Core.Class (classTyCon)
+import GHC.Unit.Env
 
 -- -----------------------------------------------------------------------------
 -- running a statement interactively
@@ -149,7 +152,7 @@ getHistoryModule = breakInfo_module . historyBreakInfo
 getHistorySpan :: HscEnv -> History -> SrcSpan
 getHistorySpan hsc_env History{..} =
   let BreakInfo{..} = historyBreakInfo in
-  case lookupHpt (hsc_HPT hsc_env) (moduleName breakInfo_module) of
+  case lookupHugByModule breakInfo_module (hsc_HUG hsc_env) of
     Just hmi -> modBreaks_locs (getModBreaks hmi) ! breakInfo_number
     _ -> panic "getHistorySpan"
 
@@ -160,7 +163,7 @@ getHistorySpan hsc_env History{..} =
 findEnclosingDecls :: HscEnv -> BreakInfo -> [String]
 findEnclosingDecls hsc_env (BreakInfo modl ix) =
    let hmi = expectJust "findEnclosingDecls" $
-             lookupHpt (hsc_HPT hsc_env) (moduleName modl)
+             lookupHugByModule modl (hsc_HUG hsc_env)
        mb = getModBreaks hmi
    in modBreaks_decls mb ! ix
 
@@ -227,11 +230,12 @@ execStmt' stmt stmt_text ExecOptions{..} = do
 
         status <-
           withVirtualCWD $
-            liftIO $
-              evalStmt interp idflags' (isStep execSingleStep) (execWrap hval)
+            liftIO $ do
+              let eval_opts = initEvalOpts idflags' (isStep execSingleStep)
+              evalStmt interp eval_opts (execWrap hval)
 
         let ic = hsc_IC hsc_env
-            bindings = (ic_tythings ic, ic_rn_gbl_env ic)
+            bindings = (ic_tythings ic, ic_gre_cache ic)
 
             size = ghciHistSize idflags'
 
@@ -308,7 +312,9 @@ emptyHistory :: Int -> BoundedList History
 emptyHistory size = nilBL size
 
 handleRunStatus :: GhcMonad m
-                => SingleStep -> String-> ([TyThing],GlobalRdrEnv) -> [Id]
+                => SingleStep -> String
+                -> ResumeBindings
+                -> [Id]
                 -> EvalStatus_ [ForeignHValue] [HValueRef]
                 -> BoundedList History
                 -> m ExecResult
@@ -342,7 +348,8 @@ handleRunStatus step expr bindings final_ids status history
                !history' = mkHistory hsc_env apStack_fhv bi `consBL` history
                  -- history is strict, otherwise our BoundedList is pointless.
            fhv <- liftIO $ mkFinalizedHValue interp resume_ctxt
-           status <- liftIO $ GHCi.resumeStmt interp dflags True fhv
+           let eval_opts = initEvalOpts dflags True
+           status <- liftIO $ GHCi.resumeStmt interp eval_opts fhv
            handleRunStatus RunAndLogSteps expr bindings final_ids
                            status history'
     | otherwise
@@ -415,9 +422,9 @@ resumeExec canLogSpan step mbCnt
         -- unbind the temporary locals by restoring the TypeEnv from
         -- before the breakpoint, and drop this Resume from the
         -- InteractiveContext.
-        let (resume_tmp_te,resume_rdr_env) = resumeBindings r
+        let (resume_tmp_te,resume_gre_cache) = resumeBindings r
             ic' = ic { ic_tythings = resume_tmp_te,
-                       ic_rn_gbl_env = resume_rdr_env,
+                       ic_gre_cache = resume_gre_cache,
                        ic_resume   = rs }
         setSession hsc_env{ hsc_IC = ic' }
 
@@ -442,7 +449,8 @@ resumeExec canLogSpan step mbCnt
                   setupBreakpoint hsc_env (fromJust mb_brkpt) (fromJust mbCnt)
                     -- When the user specified a break ignore count, set it
                     -- in the interpreter
-                status <- liftIO $ GHCi.resumeStmt interp dflags (isStep step) fhv
+                let eval_opts = initEvalOpts dflags (isStep step)
+                status <- liftIO $ GHCi.resumeStmt interp eval_opts fhv
                 let prevHistoryLst = fromListBL 50 hist
                     hist' = case mb_brkpt of
                        Nothing -> prevHistoryLst
@@ -566,7 +574,7 @@ bindLocalsAtBreakpoint hsc_env apStack_fhv (Just BreakInfo{..}) = do
 
        (ids, offsets, occs') = syncOccs mbPointers occs
 
-       free_tvs = tyCoVarsOfTypesList (result_ty:map idType ids)
+       free_tvs = tyCoVarsOfTypesWellScoped (result_ty:map idType ids)
 
    -- It might be that getIdValFromApStack fails, because the AP_STACK
    -- has been accidentally evaluated, or something else has gone wrong.
@@ -575,7 +583,7 @@ bindLocalsAtBreakpoint hsc_env apStack_fhv (Just BreakInfo{..}) = do
    mb_hValues <-
       mapM (getBreakpointVar interp apStack_fhv . fromIntegral) offsets
    when (any isNothing mb_hValues) $
-      debugTraceMsg (hsc_logger hsc_env) (hsc_dflags hsc_env) 1 $
+      debugTraceMsg (hsc_logger hsc_env) 1 $
           text "Warning: _result has been evaluated, some bindings have been lost"
 
    us <- mkSplitUniqSupply 'I'   -- Dodgy; will give the same uniques every time
@@ -616,10 +624,11 @@ bindLocalsAtBreakpoint hsc_env apStack_fhv (Just BreakInfo{..}) = do
    newTyVars :: UniqSupply -> [TcTyVar] -> TCvSubst
      -- Similarly, clone the type variables mentioned in the types
      -- we have here, *and* make them all RuntimeUnk tyvars
-   newTyVars us tvs
-     = mkTvSubstPrs [ (tv, mkTyVarTy (mkRuntimeUnkTyVar name (tyVarKind tv)))
-                    | (tv, uniq) <- tvs `zip` uniqsFromSupply us
-                    , let name = setNameUnique (tyVarName tv) uniq ]
+   newTyVars us tvs = foldl' new_tv emptyTCvSubst (tvs `zip` uniqsFromSupply us)
+   new_tv subst (tv,uniq) = extendTCvSubstWithClone subst tv new_tv
+    where
+     new_tv = mkRuntimeUnkTyVar (setNameUnique (tyVarName tv) uniq)
+                                (substTy subst (tyVarKind tv))
 
    isPointer id | [rep] <- typePrimRep (idType id)
                 , isGcPtrRep rep                   = True
@@ -662,12 +671,11 @@ rttiEnvironment hsc_env@HscEnv{hsc_IC=ic} = do
              Just new_ty -> do
               case improveRTTIType hsc_env old_ty new_ty of
                Nothing -> return $
-                        WARN(True, text (":print failed to calculate the "
-                                           ++ "improvement for a type")) hsc_env
+                        warnPprTrace True (":print failed to calculate the "
+                                           ++ "improvement for a type") empty hsc_env
                Just subst -> do
-                 let dflags = hsc_dflags hsc_env
                  let logger = hsc_logger hsc_env
-                 dumpIfSet_dyn logger dflags Opt_D_dump_rtti "RTTI"
+                 putDumpFileMaybe logger Opt_D_dump_rtti "RTTI"
                    FormatText
                    (fsep [text "RTTI Improvement for", ppr id, equals,
                           ppr subst])
@@ -684,7 +692,7 @@ pushResume hsc_env resume = hsc_env { hsc_IC = ictxt1 }
 
   {-
   Note [Syncing breakpoint info]
-
+  ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
   To display the values of the free variables for a single breakpoint, the
   function `GHC.Runtime.Eval.bindLocalsAtBreakpoint` pulls
   out the information from the fields `modBreaks_breakInfo` and
@@ -770,7 +778,7 @@ fromListBL bound l = BL (length l) bound l []
 --
 -- (setContext imports) sets the ic_imports field (which in turn
 -- determines what is in scope at the prompt) to 'imports', and
--- constructs the ic_rn_glb_env environment to reflect it.
+-- updates the icReaderEnv environment to reflect it.
 --
 -- We retain in scope all the things defined at the prompt, and kept
 -- in ic_tythings.  (Indeed, they shadow stuff from ic_imports.)
@@ -785,10 +793,10 @@ setContext imports
                liftIO $ throwGhcExceptionIO (formatError dflags mod err)
            Right all_env -> do {
        ; let old_ic         = hsc_IC hsc_env
-             !final_rdr_env = all_env `icExtendGblRdrEnv` ic_tythings old_ic
+             !final_gre_cache = ic_gre_cache old_ic `replaceImportEnv` all_env
        ; setSession
-         hsc_env{ hsc_IC = old_ic { ic_imports    = imports
-                                  , ic_rn_gbl_env = final_rdr_env }}}}
+         hsc_env{ hsc_IC = old_ic { ic_imports   = imports
+                                  , ic_gre_cache = final_gre_cache }}}}
   where
     formatError dflags mod err = ProgramError . showSDoc dflags $
       text "Cannot add module" <+> ppr mod <+>
@@ -853,7 +861,7 @@ getInfo allInfo name
        case mb_stuff of
          Nothing -> return Nothing
          Just (thing, fixity, cls_insts, fam_insts, docs) -> do
-           let rdr_env = ic_rn_gbl_env (hsc_IC hsc_env)
+           let rdr_env = icReaderEnv (hsc_IC hsc_env)
 
            -- Filter the instances based on whether the constituent names of their
            -- instance heads are all in scope.
@@ -862,7 +870,7 @@ getInfo allInfo name
            return (Just (thing, fixity, cls_insts', fam_insts', docs))
   where
     plausible rdr_env names
-          -- Dfun involving only names that are in ic_rn_glb_env
+          -- Dfun involving only names that are in icReaderEnv
         = allInfo
        || nameSetAll ok names
         where   -- A name is ok if it's in the rdr_env,
@@ -870,15 +878,14 @@ getInfo allInfo name
           ok n | n == name              = True
                        -- The one we looked for in the first place!
                | pretendNameIsInScope n = True
-               | isBuiltInSyntax n      = True
-               | isCTupleTyConName n    = True
+                   -- See Note [pretendNameIsInScope] in GHC.Builtin.Names
                | isExternalName n       = isJust (lookupGRE_Name rdr_env n)
                | otherwise              = True
 
 -- | Returns all names in scope in the current interactive context
 getNamesInScope :: GhcMonad m => m [Name]
 getNamesInScope = withSession $ \hsc_env ->
-  return (map greMangledName (globalRdrEnvElts (ic_rn_gbl_env (hsc_IC hsc_env))))
+  return (map greMangledName (globalRdrEnvElts (icReaderEnv (hsc_IC hsc_env))))
 
 -- | Returns all 'RdrName's in scope in the current interactive
 -- context, excluding any that are internally-generated.
@@ -886,7 +893,7 @@ getRdrNamesInScope :: GhcMonad m => m [RdrName]
 getRdrNamesInScope = withSession $ \hsc_env -> do
   let
       ic = hsc_IC hsc_env
-      gbl_rdrenv = ic_rn_gbl_env ic
+      gbl_rdrenv = icReaderEnv ic
       gbl_names = concatMap greRdrNames $ globalRdrEnvElts gbl_rdrenv
   -- Exclude internally generated names; see e.g. #11328
   return (filter (not . isDerivedOccName . rdrNameOcc) gbl_names)
@@ -902,7 +909,7 @@ parseName str = withSession $ \hsc_env -> liftIO $
 
 getDocs :: GhcMonad m
         => Name
-        -> m (Either GetDocsFailure (Maybe HsDocString, IntMap HsDocString))
+        -> m (Either GetDocsFailure (Maybe [HsDoc GhcRn], IntMap (HsDoc GhcRn)))
            -- TODO: What about docs for constructors etc.?
 getDocs name =
   withSession $ \hsc_env -> do
@@ -912,14 +919,14 @@ getDocs name =
          if isInteractiveModule mod
            then pure (Left InteractiveName)
            else do
-             ModIface { mi_doc_hdr = mb_doc_hdr
-                      , mi_decl_docs = DeclDocMap dmap
-                      , mi_arg_docs = ArgDocMap amap
-                      } <- liftIO $ hscGetModuleInterface hsc_env mod
-             if isNothing mb_doc_hdr && Map.null dmap && Map.null amap
-               then pure (Left (NoDocsInIface mod compiled))
-               else pure (Right ( Map.lookup name dmap
-                                , Map.findWithDefault mempty name amap))
+             iface <- liftIO $ hscGetModuleInterface hsc_env mod
+             case mi_docs iface of
+               Nothing -> pure (Left (NoDocsInIface mod compiled))
+               Just Docs { docs_decls = decls
+                         , docs_args = args
+                         } ->
+                 pure (Right ( lookupUniqMap decls name
+                             , fromMaybe mempty $ lookupUniqMap args name))
   where
     compiled =
       -- TODO: Find a more direct indicator.
@@ -928,16 +935,12 @@ getDocs name =
         UnhelpfulLoc {} -> True
 
 -- | Failure modes for 'getDocs'.
-
--- TODO: Find a way to differentiate between modules loaded without '-haddock'
--- and modules that contain no docs.
 data GetDocsFailure
 
     -- | 'nameModule_maybe' returned 'Nothing'.
   = NameHasNoModule Name
 
-    -- | This is probably because the module was loaded without @-haddock@,
-    -- but it's also possible that the entire module contains no documentation.
+    -- | The module was loaded without @-haddock@,
   | NoDocsInIface
       Module
       Bool -- ^ 'True': The module was compiled.
@@ -951,11 +954,6 @@ instance Outputable GetDocsFailure where
     quotes (ppr name) <+> text "has no module where we could look for docs."
   ppr (NoDocsInIface mod compiled) = vcat
     [ text "Can't find any documentation for" <+> ppr mod <> char '.'
-    , text "This is probably because the module was"
-        <+> text (if compiled then "compiled" else "loaded")
-        <+> text "without '-haddock',"
-    , text "but it's also possible that the module contains no documentation."
-    , text ""
     , if compiled
         then text "Try re-compiling with '-haddock'."
         else text "Try running ':set -haddock' and :load the file again."
@@ -1032,7 +1030,7 @@ typeKind normalise str = withSession $ \hsc_env ->
 getInstancesForType :: GhcMonad m => Type -> m [ClsInst]
 getInstancesForType ty = withSession $ \hsc_env ->
   liftIO $ runInteractiveHsc hsc_env $
-    ioMsgMaybe $ runTcInteractive hsc_env $ do
+    ioMsgMaybe $ hoistTcRnMessage $ runTcInteractive hsc_env $ do
       -- Bring class and instances from unqualified modules into scope, this fixes #16793.
       loadUnqualIfaces hsc_env (hsc_IC hsc_env)
       matches <- findMatchingInstances ty
@@ -1045,7 +1043,7 @@ parseInstanceHead str = withSession $ \hsc_env0 -> do
   (ty, _) <- liftIO $ runInteractiveHsc hsc_env0 $ do
     hsc_env <- getHscEnv
     ty <- hscParseType str
-    ioMsgMaybe $ tcRnType hsc_env SkolemiseFlexi True ty
+    ioMsgMaybe $ hoistTcRnMessage $ tcRnType hsc_env SkolemiseFlexi True ty
 
   return ty
 
@@ -1054,24 +1052,20 @@ getDictionaryBindings :: PredType -> TcM CtEvidence
 getDictionaryBindings theta = do
   dictName <- newName (mkDictOcc (mkVarOcc "magic"))
   let dict_var = mkVanillaGlobal dictName theta
-  loc <- getCtLocM (GivenOrigin UnkSkol) Nothing
+  loc <- getCtLocM (GivenOrigin (getSkolemInfo unkSkol)) Nothing
 
-  -- Generate a wanted here because at the end of constraint
-  -- solving, most derived constraints get thrown away, which in certain
-  -- cases, notably with quantified constraints makes it impossible to rule
-  -- out instances as invalid. (See #18071)
   return CtWanted {
     ctev_pred = varType dict_var,
     ctev_dest = EvVarDest dict_var,
-    ctev_nosh = WDeriv,
-    ctev_loc = loc
+    ctev_loc = loc,
+    ctev_rewriters = emptyRewriterSet
   }
 
 -- Find instances where the head unifies with the provided type
 findMatchingInstances :: Type -> TcM [(ClsInst, [DFunInstType])]
 findMatchingInstances ty = do
   ies@(InstEnvs {ie_global = ie_global, ie_local = ie_local}) <- tcGetInstEnvs
-  let allClasses = instEnvClasses ie_global ++ instEnvClasses ie_local
+  let allClasses = uniqDSetToList $ instEnvClasses ie_global `unionUniqDSets` instEnvClasses ie_local
   return $ concatMap (try_cls ies) allClasses
   where
   {- Check that a class instance is well-kinded.
@@ -1125,13 +1119,19 @@ checkForExistence clsInst mb_inst_tys = do
   -- which otherwise appear as opaque type variables. (See #18262).
   WC { wc_simple = simples, wc_impl = impls } <- simplifyWantedsTcM wanteds
 
-  if allBag allowedSimple simples && solvedImplics impls
-  then return . Just $ substInstArgs tys (bagToList (mapBag ctPred simples)) clsInst
+  -- The simples might contain superclasses. This clutters up the output
+  -- (we want e.g. instance Ord a => Ord (Maybe a), not
+  -- instance (Ord a, Eq a) => Ord (Maybe a)). So we use mkMinimalBySCs
+  let simple_preds = map ctPred (bagToList simples)
+  let minimal_simples = mkMinimalBySCs id simple_preds
+
+  if all allowedSimple minimal_simples && solvedImplics impls
+  then return . Just $ substInstArgs tys minimal_simples clsInst
   else return Nothing
 
   where
-  allowedSimple :: Ct -> Bool
-  allowedSimple ct = isSatisfiablePred (ctPred ct)
+  allowedSimple :: PredType -> Bool
+  allowedSimple pred = isSatisfiablePred pred
 
   solvedImplics :: Bag Implication -> Bool
   solvedImplics impls = allBag (isSolvedStatus . ic_status) impls
@@ -1211,7 +1211,8 @@ compileParsedExprRemote expr@(L loc _) = withSession $ \hsc_env -> do
         _ -> panic "compileParsedExprRemote"
 
   updateFixityEnv fix_env
-  status <- liftIO $ evalStmt interp dflags False (EvalThis hvals_io)
+  let eval_opts = initEvalOpts dflags False
+  status <- liftIO $ evalStmt interp eval_opts (EvalThis hvals_io)
   case status of
     EvalComplete _ (EvalSuccess [hval]) -> return hval
     EvalComplete _ (EvalException e) ->
@@ -1243,8 +1244,7 @@ showModule mod_summary =
     withSession $ \hsc_env -> do
         interpreted <- moduleIsBootOrNotObjectLinkable mod_summary
         let dflags = hsc_dflags hsc_env
-        -- extendModSummaryNoDeps because the message doesn't look at the deps
-        return (showSDoc dflags $ showModMsg dflags interpreted (ModuleNode (extendModSummaryNoDeps mod_summary)))
+        return (showSDoc dflags $ showModMsg dflags interpreted (ModuleNode [] mod_summary))
 
 moduleIsBootOrNotObjectLinkable :: GhcMonad m => ModSummary -> m Bool
 moduleIsBootOrNotObjectLinkable mod_summary = withSession $ \hsc_env ->
@@ -1271,13 +1271,13 @@ obtainTermFromVal hsc_env _bound _force _ty _x = case interpInstance interp of
 
 obtainTermFromId :: HscEnv -> Int -> Bool -> Id -> IO Term
 obtainTermFromId hsc_env bound force id =  do
-  hv <- Loader.loadName (hscInterp hsc_env) hsc_env (varName id)
+  (hv, _, _) <- Loader.loadName (hscInterp hsc_env) hsc_env (varName id)
   cvObtainTerm hsc_env bound force (idType id) hv
 
 -- Uses RTTI to reconstruct the type of an Id, making it less polymorphic
 reconstructType :: HscEnv -> Int -> Id -> IO (Maybe Type)
 reconstructType hsc_env bound id = do
-  hv <- Loader.loadName (hscInterp hsc_env) hsc_env (varName id)
+  (hv, _, _) <- Loader.loadName (hscInterp hsc_env) hsc_env (varName id)
   cvReconstructType hsc_env bound (idType id) hv
 
 mkRuntimeUnkTyVar :: Name -> Kind -> TyVar
