@@ -1,4 +1,4 @@
-{-# LANGUAGE CPP #-}
+
 {-# LANGUAGE DeriveFunctor #-}
 {-# LANGUAGE TypeFamilies #-}
 
@@ -27,10 +27,9 @@ module GHC.Tc.TyCl.Utils(
         tcRecSelBinds, mkRecSelBinds, mkOneRecordSelector
     ) where
 
-#include "HsVersions.h"
-
 import GHC.Prelude
 
+import GHC.Tc.Errors.Types
 import GHC.Tc.Utils.Monad
 import GHC.Tc.Utils.Env
 import GHC.Tc.Gen.Bind( tcValBinds )
@@ -55,6 +54,7 @@ import GHC.Core.Coercion ( ltRole )
 
 import GHC.Utils.Outputable
 import GHC.Utils.Panic
+import GHC.Utils.Panic.Plain
 import GHC.Utils.Misc
 import GHC.Utils.FV as FV
 
@@ -65,6 +65,7 @@ import GHC.Data.FastString
 import GHC.Unit.Module
 
 import GHC.Types.Basic
+import GHC.Types.Error
 import GHC.Types.FieldLabel
 import GHC.Types.SrcLoc
 import GHC.Types.SourceFile
@@ -91,11 +92,11 @@ import Control.Monad
 -}
 
 synonymTyConsOfType :: Type -> [TyCon]
--- Does not look through type synonyms at all
--- Return a list of synonym tycons
+-- Does not look through type synonyms at all.
+-- Returns a list of synonym tycons in nondeterministic order.
 -- Keep this synchronized with 'expandTypeSynonyms'
 synonymTyConsOfType ty
-  = nameEnvElts (go ty)
+  = nonDetNameEnvElts (go ty)
   where
      go :: Type -> NameEnv TyCon  -- The NameEnv does duplicate elim
      go (TyConApp tc tys) = go_tc tc `plusNameEnv` go_s tys
@@ -205,7 +206,7 @@ checkTyConIsAcyclic tc m = SynCycleM $ \s ->
 checkSynCycles :: Unit -> [TyCon] -> [LTyClDecl GhcRn] -> TcM ()
 checkSynCycles this_uid tcs tyclds =
     case runSynCycleM (mapM_ (go emptyTyConSet []) tcs) emptyTyConSet of
-        Left (loc, err) -> setSrcSpan loc $ failWithTc err
+        Left (loc, err) -> setSrcSpan loc $ failWithTc (TcRnUnknownMessage $ mkPlainError noHints err)
         Right _  -> return ()
   where
     -- Try our best to print the LTyClDecl for locally defined things
@@ -650,6 +651,8 @@ markNominal lcls ty = let nvars = fvVarList (FV.delFVs lcls $ get_ty_vars ty) in
      -- recurring into coercions. Recall: coercions are totally ignored during
      -- role inference. See [Coercions in role inference]
     get_ty_vars :: Type -> FV
+    get_ty_vars t                 | Just t' <- coreView t -- #20999
+                                  = get_ty_vars t'
     get_ty_vars (TyVarTy tv)      = unitFV tv
     get_ty_vars (AppTy t1 t2)     = get_ty_vars t1 `unionFV` get_ty_vars t2
     get_ty_vars (FunTy _ w t1 t2) = get_ty_vars w `unionFV` get_ty_vars t1 `unionFV` get_ty_vars t2
@@ -715,21 +718,21 @@ runRoleM env thing = (env', update)
 
 setRoleInferenceTc :: Name -> RoleM a -> RoleM a
 setRoleInferenceTc name thing = RM $ \m_name vps nvps state ->
-                                ASSERT( isNothing m_name )
-                                ASSERT( isEmptyVarEnv vps )
-                                ASSERT( nvps == 0 )
+                                assert (isNothing m_name) $
+                                assert (isEmptyVarEnv vps) $
+                                assert (nvps == 0) $
                                 unRM thing (Just name) vps nvps state
 
 addRoleInferenceVar :: TyVar -> RoleM a -> RoleM a
 addRoleInferenceVar tv thing
   = RM $ \m_name vps nvps state ->
-    ASSERT( isJust m_name )
+    assert (isJust m_name) $
     unRM thing m_name (extendVarEnv vps tv nvps) (nvps+1) state
 
 setRoleInferenceVars :: [TyVar] -> RoleM a -> RoleM a
 setRoleInferenceVars tvs thing
   = RM $ \m_name _vps _nvps state ->
-    ASSERT( isJust m_name )
+    assert (isJust m_name) $
     unRM thing m_name (mkVarEnv (zip tvs [0..])) (panic "setRoleInferenceVars")
          state
 
@@ -764,12 +767,14 @@ updateRoleEnv name n role
 *                                                                      *
 ********************************************************************* -}
 
-addTyConsToGblEnv :: [TyCon] -> TcM TcGblEnv
+addTyConsToGblEnv :: [TyCon] -> TcM (TcGblEnv, ThBindEnv)
 -- Given a [TyCon], add to the TcGblEnv
 --   * extend the TypeEnv with the tycons
 --   * extend the TypeEnv with their implicitTyThings
 --   * extend the TypeEnv with any default method Ids
 --   * add bindings for record selectors
+-- Return separately the TH levels of these bindings,
+-- to be added to a LclEnv later.
 addTyConsToGblEnv tyclss
   = tcExtendTyConEnv tyclss                    $
     tcExtendGlobalEnvImplicit implicit_things  $
@@ -777,7 +782,10 @@ addTyConsToGblEnv tyclss
     do { traceTc "tcAddTyCons" $ vcat
             [ text "tycons" <+> ppr tyclss
             , text "implicits" <+> ppr implicit_things ]
-       ; tcRecSelBinds (mkRecSelBinds tyclss) }
+       ; gbl_env <- tcRecSelBinds (mkRecSelBinds tyclss)
+       ; th_bndrs <- tcTyThBinders implicit_things
+       ; return (gbl_env, th_bndrs)
+       }
  where
    implicit_things = concatMap implicitTyConThings tyclss
    def_meth_ids    = mkDefaultMethodIds tyclss
@@ -880,6 +888,7 @@ mkOneRecordSelector all_cons idDetails fl has_sel
     loc      = getSrcSpan sel_name
     loc'     = noAnnSrcSpan loc
     locn     = noAnnSrcSpan loc
+    locc     = noAnnSrcSpan loc
     lbl      = flLabel fl
     sel_name = flSelector fl
 
@@ -888,7 +897,7 @@ mkOneRecordSelector all_cons idDetails fl has_sel
 
     -- Find a representative constructor, con1
     cons_w_field = conLikesWithFields all_cons [lbl]
-    con1 = ASSERT( not (null cons_w_field) ) head cons_w_field
+    con1 = assert (not (null cons_w_field)) $ head cons_w_field
 
     -- Selector type; Note [Polymorphic selectors]
     field_ty   = conLikeFieldType con1 lbl
@@ -899,7 +908,8 @@ mkOneRecordSelector all_cons idDetails fl has_sel
                     || has_sel == NoFieldSelectors
     sel_ty | is_naughty = unitTy  -- See Note [Naughty record selectors]
            | otherwise  = mkForAllTys (tyVarSpecToBinders data_tvbs) $
-                          mkPhiTy (conLikeStupidTheta con1) $   -- Urgh!
+                          -- Urgh! See Note [The stupid context] in GHC.Core.DataCon
+                          mkPhiTy (conLikeStupidTheta con1) $
                           -- req_theta is empty for normal DataCon
                           mkPhiTy req_theta                 $
                           mkVisFunTyMany data_ty            $
@@ -921,14 +931,14 @@ mkOneRecordSelector all_cons idDetails fl has_sel
                                  (L loc' (HsVar noExtField (L locn field_var)))
     mk_sel_pat con = ConPat NoExtField (L locn (getName con)) (RecCon rec_fields)
     rec_fields = HsRecFields { rec_flds = [rec_field], rec_dotdot = Nothing }
-    rec_field  = noLocA (HsRecField
-                        { hsRecFieldAnn = noAnn
-                        , hsRecFieldLbl
-                           = L loc (FieldOcc sel_name
-                                     (L locn $ mkVarUnqual lbl))
-                        , hsRecFieldArg
+    rec_field  = noLocA (HsFieldBind
+                        { hfbAnn = noAnn
+                        , hfbLHS
+                           = L locc (FieldOcc sel_name
+                                      (L locn $ mkVarUnqual lbl))
+                        , hfbRHS
                            = L loc' (VarPat noExtField (L locn field_var))
-                        , hsRecPun = False })
+                        , hfbPun = False })
     sel_lname = L locn sel_name
     field_var = mkInternalName (mkBuiltinUnique 1) (getOccName sel_name) loc
 
@@ -961,14 +971,14 @@ mkOneRecordSelector all_cons idDetails fl has_sel
     eq_subst = mkTvSubstPrs (map eqSpecPair eq_spec)
     -- inst_tys corresponds to one of the following:
     --
-    -- * The arguments to the user-written return type (for GADT constructors).
-    --   In this scenario, eq_subst provides a mapping from the universally
-    --   quantified type variables to the argument types. Note that eq_subst
-    --   does not need to be applied to any other part of the DataCon
-    --   (see Note [The dcEqSpec domain invariant] in GHC.Core.DataCon).
-    -- * The universally quantified type variables
-    --   (for Haskell98-style constructors and pattern synonyms). In these
-    --   scenarios, eq_subst is an empty substitution.
+    --  * The arguments to the user-written return type (for GADT constructors).
+    --    In this scenario, eq_subst provides a mapping from the universally
+    --    quantified type variables to the argument types. Note that eq_subst
+    --    does not need to be applied to any other part of the DataCon
+    --    (see Note [The dcEqSpec domain invariant] in GHC.Core.DataCon).
+    --  * The universally quantified type variables
+    --    (for Haskell98-style constructors and pattern synonyms). In these
+    --    scenarios, eq_subst is an empty substitution.
     inst_tys = substTyVars eq_subst univ_tvs
 
     unit_rhs = mkLHsTupleExpr [] noExtField
