@@ -12,7 +12,9 @@ packageArgs :: Args
 packageArgs = do
     stage        <- getStage
     path         <- getBuildPath
+    root         <- getBuildRoot
     compilerPath <- expr $ buildPath (vanillaContext stage compiler)
+
     let -- Do not bind the result to a Boolean: this forces the configure rule
         -- immediately and may lead to cyclic dependencies.
         -- See: https://gitlab.haskell.org/ghc/ghc/issues/16809.
@@ -20,12 +22,13 @@ packageArgs = do
 
         -- Check if the bootstrap compiler has the same version as the one we
         -- are building. This is used to build cross-compilers
-        bootCross = (==) <$> ghcVersionStage Stage0 <*> ghcVersionStage Stage1
+        bootCross = (==) <$> ghcVersionStage (stage0InTree) <*> ghcVersionStage Stage1
 
     cursesIncludeDir <- getSetting CursesIncludeDir
     cursesLibraryDir <- getSetting CursesLibDir
     ffiIncludeDir  <- getSetting FfiIncludeDir
     ffiLibraryDir  <- getSetting FfiLibDir
+    debugAssertions  <- ghcDebugAssertions <$> expr flavour
 
     mconcat
         --------------------------------- base ---------------------------------
@@ -48,9 +51,14 @@ packageArgs = do
           [ builder Alex ? arg "--latin1"
 
           , builder (Ghc CompileHs) ? mconcat
-            [ inputs ["**/GHC.hs", "**/GHC/Driver/Make.hs"] ? arg "-fprof-auto"
+            [ debugAssertions ? notStage0 ? arg "-DDEBUG"
+
+            , inputs ["**/GHC.hs", "**/GHC/Driver/Make.hs"] ? arg "-fprof-auto"
             , input "**/Parser.hs" ?
               pure ["-fno-ignore-interface-pragmas", "-fcmm-sink"]
+            -- Enable -haddock and -Winvalid-haddock for the compiler
+            , arg "-haddock"
+            , notStage0 ? arg "-Winvalid-haddock"
             -- These files take a very long time to compile with -O1,
             -- so we use -O0 for them just in Stage0 to speed up the
             -- build but not affect Stage1+ executables
@@ -60,25 +68,22 @@ packageArgs = do
           , builder (Cabal Setup) ? mconcat
             [ arg "--disable-library-for-ghci"
             , anyTargetOs ["openbsd"] ? arg "--ld-options=-E"
-            , flag GhcUnregisterised ? arg "--ghc-option=-DNO_REGS"
-            , notM targetSupportsSMP ? arg "--ghc-option=-DNOSMP"
-            , notM targetSupportsSMP ? arg "--ghc-option=-optc-DNOSMP"
-            , ghcWithInterpreter ?
-              ghciWithDebugger <$> flavour ?
-              notStage0 ? arg "--ghc-option=-DDEBUGGER"
             , ghcProfiled <$> flavour ?
               notStage0 ? arg "--ghc-pkg-option=--force" ]
 
           , builder (Cabal Flags) ? mconcat
             [ andM [expr ghcWithInterpreter, notStage0] `cabalFlag` "internal-interpreter"
             , notM cross `cabalFlag` "terminfo"
+            , arg "-build-tool-depends"
             ]
 
           , builder (Haddock BuildPackage) ? arg ("--optghc=-I" ++ path) ]
 
         ---------------------------------- ghc ---------------------------------
         , package ghc ? mconcat
-          [ builder Ghc ? arg ("-I" ++ compilerPath)
+          [ builder Ghc ? mconcat
+             [ arg ("-I" ++ compilerPath)
+             , debugAssertions ? notStage0 ? arg "-DDEBUG" ]
 
           , builder (Cabal Flags) ? mconcat
             [ andM [expr ghcWithInterpreter, notStage0] `cabalFlag` "internal-interpreter"
@@ -158,6 +163,14 @@ packageArgs = do
         , package haddock ?
           builder (Cabal Flags) ? arg "in-ghc-tree"
 
+        ---------------------------------- text --------------------------------
+        , package text ? mconcat
+          -- Disable SIMDUTF by default due to packaging difficulties
+          -- described in #20724.
+          [ builder (Cabal Flags) ? arg "-simdutf"
+          -- https://github.com/haskell/text/issues/415
+          , builder Ghc ? input "**/Data/Text/Encoding.hs"  ? arg "-Wno-unused-imports" ]
+
         ------------------------------- haskeline ------------------------------
         -- Hadrian doesn't currently support packages containing both libraries
         -- and executables. This flag disables the latter.
@@ -177,6 +190,10 @@ packageArgs = do
         , package hsc2hs ?
           builder (Cabal Flags) ? arg "in-ghc-tree"
 
+        -------------------------------- genapply --------------------------------
+        -- TODO: The logic here needs to come first, so it's hacked into
+        -- Settings.Builder.Ghc instead.
+
         ------------------------------ ghc-bignum ------------------------------
         , ghcBignumArgs
 
@@ -187,6 +204,20 @@ packageArgs = do
         , package runGhc ?
           builder Ghc ? input "**/Main.hs" ?
           (\version -> ["-cpp", "-DVERSION=" ++ show version]) <$> getSetting ProjectVersion
+
+        --------------------------------- genprimopcode ------------------------
+        , package genprimopcode
+          ? builder (Cabal Flags) ? arg "-build-tool-depends"
+
+        --------------------------------- hpcBin ----------------------------------
+        , package hpcBin
+          ? builder (Cabal Flags) ? arg "-build-tool-depends"
+
+        --------------------------------- template-haskell ----------------------------------
+
+        , package templateHaskell
+            ? mconcat [ builder (Cabal Flags) ? notStage0 ? arg "+vendor-filepath"
+                      , builder Ghc ? notStage0 ? arg ("-i" <> (root </> pkgPath filepath)) ]
         ]
 
 ghcBignumArgs :: Args
@@ -352,18 +383,12 @@ rtsPackageArgs = package rts ? do
 
             , input "**/RetainerProfile.c" ? flag CcLlvmBackend ?
               arg "-Wno-incompatible-pointer-types"
-
-            -- libffi's ffi.h triggers various warnings
-            , inputs [ "**/Interpreter.c", "**/Storage.c", "**/Adjustor.c" ] ?
-              arg "-Wno-strict-prototypes"
-            , inputs ["**/Interpreter.c", "**/Adjustor.c", "**/sm/Storage.c"] ?
-              anyTargetArch ["powerpc"] ? arg "-Wno-undef" ]
+            ]
 
     mconcat
         [ builder (Cabal Flags) ? mconcat
           [ any (wayUnit Profiling) rtsWays `cabalFlag` "profiling"
           , any (wayUnit Debug) rtsWays     `cabalFlag` "debug"
-          , any (wayUnit Logging) rtsWays   `cabalFlag` "logging"
           , any (wayUnit Dynamic) rtsWays   `cabalFlag` "dynamic"
           , useSystemFfi                    `cabalFlag` "use-system-libffi"
           , useLibffiForAdjustors           `cabalFlag` "libffi-adjustors"
@@ -375,7 +400,7 @@ rtsPackageArgs = package rts ? do
               , useSystemFfi ? cabalExtraDirs ffiIncludeDir ffiLibraryDir
               ]
         , builder (Cc (FindCDependencies CDep)) ? cArgs
-        , builder (Cc (FindCDependencies CxxDep)) ? cArgs
+        , builder (Cc (FindCDependencies  CxxDep)) ? cArgs
         , builder (Ghc CompileCWithGhc) ? map ("-optc" ++) <$> cArgs
         , builder (Ghc CompileCppWithGhc) ? map ("-optcxx" ++) <$> cArgs
         , builder Ghc ? ghcArgs
@@ -401,7 +426,7 @@ rtsPackageArgs = package rts ? do
 -- On Darwin we get errors of the form
 --
 --  ld: absolute addressing (perhaps -mdynamic-no-pic) used in _stg_ap_0_fast
---      from rts/dist/build/Apply.dyn_o not allowed in slidable image
+--      from rts/dist-install/build/Apply.dyn_o not allowed in slidable image
 --
 -- and lots of these warnings:
 --
@@ -412,7 +437,7 @@ rtsPackageArgs = package rts ? do
 --
 -- Text relocation remains                         referenced
 --     against symbol                  offset      in file
--- .rodata (section)                   0x11        rts/dist/build/Apply.dyn_o
+-- .rodata (section)                   0x11        rts/dist-install/build/Apply.dyn_o
 --   ...
 -- ld: fatal: relocations remain against allocatable but non-writable sections
 -- collect2: ld returned 1 exit status
