@@ -1,4 +1,7 @@
-{-# LANGUAGE Safe #-}
+{-# LANGUAGE FlexibleContexts     #-}
+{-# LANGUAGE InstanceSigs         #-}
+{-# LANGUAGE Safe                 #-}
+{-# LANGUAGE UndecidableInstances #-}
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Data.Functor.Classes
@@ -6,7 +9,7 @@
 -- License     :  BSD-style (see the file LICENSE)
 --
 -- Maintainer  :  libraries@haskell.org
--- Stability   :  experimental
+-- Stability   :  stable
 -- Portability :  portable
 --
 -- Liftings of the Prelude classes 'Eq', 'Ord', 'Read' and 'Show' to
@@ -72,6 +75,7 @@ import Data.List.NonEmpty (NonEmpty(..))
 import Data.Ord (Down(Down))
 import Data.Complex (Complex((:+)))
 
+import GHC.Generics (Generic1(..), Generically1(..))
 import GHC.Tuple (Solo (..))
 import GHC.Read (expectP, list, paren)
 
@@ -688,6 +692,15 @@ instance (Read a, Read b, Read c) => Read1 ((,,,) a b c) where
 instance (Show a, Show b, Show c) => Show1 ((,,,) a b c) where
     liftShowsPrec = liftShowsPrec2 showsPrec showList
 
+-- | @since 4.17.0.0
+instance (Generic1 f, Eq1 (Rep1 f)) => Eq1 (Generically1 f) where
+  liftEq :: (a1 -> a2 -> Bool) -> (Generically1 f a1 -> Generically1 f a2 -> Bool)
+  liftEq (===) (Generically1 as1) (Generically1 as2) = liftEq (===) (from1 as1) (from1 as2)
+
+-- | @since 4.17.0.0
+instance (Generic1 f, Ord1 (Rep1 f)) => Ord1 (Generically1 f) where
+  liftCompare :: (a1 -> a2 -> Ordering) -> (Generically1 f a1 -> Generically1 f a2 -> Ordering)
+  liftCompare cmp (Generically1 as1) (Generically1 as2) = liftCompare cmp (from1 as1) (from1 as2)
 
 -- | @since 4.9.0.0
 instance Eq2 Either where
@@ -821,7 +834,10 @@ instance Eq1 Down where
 
 -- | @since 4.12.0.0
 instance Ord1 Down where
-    liftCompare comp (Down x) (Down y) = comp x y
+    liftCompare comp (Down x) (Down y) = case comp x y of
+        LT -> GT
+        EQ -> EQ
+        GT -> LT
 
 -- | @since 4.12.0.0
 instance Read1 Down where
