@@ -1,13 +1,7 @@
-{-# LANGUAGE CPP, BangPatterns, ScopedTypeVariables #-}
+{-# LANGUAGE BangPatterns, ScopedTypeVariables #-}
 {-# LANGUAGE MagicHash, UnboxedTuples, PatternGuards #-}
 {-# OPTIONS_GHC -fno-warn-unused-imports #-}
-#if __GLASGOW_HASKELL__ == 700
--- This is needed as a workaround for an old bug in GHC 7.0.1 (Trac #4498)
-{-# LANGUAGE MonoPatBinds #-}
-#endif
-#if __GLASGOW_HASKELL__ >= 701
 {-# LANGUAGE Trustworthy #-}
-#endif
 {- | Copyright : (c) 2010-2011 Simon Meier
                  (c) 2010      Jasper van der Jeugt
 License        : BSD3-style (see LICENSE)
@@ -445,7 +439,7 @@ module Data.ByteString.Builder.Prim (
 {-
   -- * Testing support
   -- | The following four functions are intended for testing use
-  -- only. They are /not/ efficient. Basic encodings are efficently executed by
+  -- only. They are /not/ efficient. Basic encodings are efficiently executed by
   -- creating 'Builder's from them using the @encodeXXX@ functions explained at
   -- the top of this module.
 
@@ -472,17 +466,9 @@ import qualified Data.ByteString.Builder.Prim.Internal as I (size, sizeBound)
 import           Data.ByteString.Builder.Prim.Binary
 import           Data.ByteString.Builder.Prim.ASCII
 
-#if MIN_VERSION_base(4,4,0)
-#if MIN_VERSION_base(4,7,0)
 import           Foreign
 import           Foreign.C.Types
-#else
-import           Foreign hiding (unsafeForeignPtrToPtr)
-#endif
 import           Foreign.ForeignPtr.Unsafe (unsafeForeignPtrToPtr)
-#else
-import           Foreign
-#endif
 import           GHC.Word (Word8 (..))
 import           GHC.Exts
 import           GHC.IO
@@ -546,9 +532,9 @@ primMapLazyByteStringFixed = primMapLazyByteStringBounded . toB
 -- influences the boundaries of the generated chunks. However, for a user of
 -- this library it is observationally equivalent, as chunk boundaries of a lazy
 -- 'L.ByteString' can only be observed through the internal interface.
--- Morevoer, we expect that all primitives write much fewer than 4kb (the
+-- Moreover, we expect that all primitives write much fewer than 4kb (the
 -- default short buffer size). Hence, it is safe to ignore the additional
--- memory spilled due to the more agressive buffer wrapping introduced by this
+-- memory spilled due to the more aggressive buffer wrapping introduced by this
 -- optimization.
 --
 {-# INLINE[1] primBounded #-}
@@ -681,13 +667,10 @@ primMapLazyByteStringBounded w =
 -- Raw CString encoding
 ------------------------------------------------------------------------------
 
-#if !MIN_VERSION_base(4,7,0)
--- eqWord# et al. return Bools prior to GHC 7.6
-isTrue# :: Bool -> Bool
-isTrue# x = x
-#endif
-
--- | A null-terminated ASCII encoded 'CString'. Null characters are not representable.
+-- | A null-terminated ASCII encoded 'Foreign.C.String.CString'.
+-- Null characters are not representable.
+--
+-- @since 0.11.0.0
 cstring :: Addr# -> Builder
 cstring =
     \addr0 -> builder $ step addr0
@@ -705,8 +688,10 @@ cstring =
       where
         !ch = indexWord8OffAddr# addr 0#
 
--- | A null-terminated UTF-8 encoded 'CString'. Null characters can be encoded as
--- @0xc0 0x80@.
+-- | A null-terminated UTF-8 encoded 'Foreign.C.String.CString'.
+-- Null characters can be encoded as @0xc0 0x80@.
+--
+-- @since 0.11.0.0
 cstringUtf8 :: Addr# -> Builder
 cstringUtf8 =
     \addr0 -> builder $ step addr0
