@@ -129,8 +129,6 @@ module Data.ByteString.Char8 (
         groupBy,
         inits,
         tails,
-        initsNE,
-        tailsNE,
         strip,
         stripPrefix,
         stripSuffix,
@@ -251,7 +249,7 @@ import qualified Data.ByteString.Unsafe as B
 
 -- Listy functions transparently exported
 import Data.ByteString (null,length,tail,init,append
-                       ,inits,tails,initsNE,tailsNE,reverse,transpose
+                       ,inits,tails,reverse,transpose
                        ,concat,take,takeEnd,drop,dropEnd,splitAt
                        ,intercalate,sort,isPrefixOf,isSuffixOf
                        ,isInfixOf,stripPrefix,stripSuffix
@@ -265,7 +263,7 @@ import Data.ByteString (null,length,tail,init,append
                        ,useAsCString,useAsCStringLen
                        )
 
-import Data.ByteString.Internal.Type
+import Data.ByteString.Internal
 
 import Data.Char    ( isSpace )
 -- See bytestring #70
@@ -971,11 +969,11 @@ unlines = \li -> let
   (+!) = checkedAdd "Char8.unlines"
 
   go [] _ = pure ()
-  go (BS src len : srcs) dest = do
-    memcpyFp dest src len
-    pokeFpByteOff dest len (c2w '\n')
-    go srcs $ dest `plusForeignPtr` (len + 1)
-  in  unsafeCreateFp totLen (go li)
+  go (BS srcFP len : srcs) dest = do
+    unsafeWithForeignPtr srcFP $ \src -> memcpy dest src len
+    pokeElemOff dest len (c2w '\n')
+    go srcs $ dest `plusPtr` (len + 1)
+  in  unsafeCreate totLen (go li)
 
 -- | 'words' breaks a ByteString up into a list of words, which
 -- were delimited by Chars representing white space.
@@ -1067,20 +1065,12 @@ readInteger as
 ------------------------------------------------------------------------
 -- For non-binary text processing:
 
--- | Write a ByteString to a handle, appending a newline byte.
---
--- Unlike 'hPutStr', this is not atomic: other threads might write
--- to the handle between writing of the bytestring and the newline.
---
+-- | Write a ByteString to a handle, appending a newline byte
 hPutStrLn :: Handle -> ByteString -> IO ()
 hPutStrLn h ps
     | length ps < 1024 = hPut h (ps `B.snoc` 0x0a)
     | otherwise        = hPut h ps >> hPut h (B.singleton 0x0a) -- don't copy
 
--- | Write a ByteString to 'stdout', appending a newline byte.
---
--- Unlike 'putStr', this is not atomic: other threads might write
--- to 'stdout' between writing of the bytestring and the newline.
---
+-- | Write a ByteString to stdout, appending a newline byte
 putStrLn :: ByteString -> IO ()
 putStrLn = hPutStrLn stdout
