@@ -5,7 +5,7 @@
  * ---------------------------------------------------------------------------*/
 
 
-#include "PosixSource.h"
+#include "rts/PosixSource.h"
 #include "Rts.h"
 #include "RtsAPI.h"
 #include "rts/Bytecodes.h"
@@ -103,7 +103,7 @@
 #endif
 
 // Note [Not true: ASSERT(Sp > SpLim)]
-//
+// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 // SpLim has some headroom (RESERVED_STACK_WORDS) to allow for saving
 // any necessary state on the stack when returning to the scheduler
 // when a stack check fails..  The upshot of this is that Sp could be
@@ -117,7 +117,7 @@
    return cap;
 
 // Note [avoiding threadPaused]
-//
+// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 // Switching between the interpreter to compiled code can happen very
 // frequently, so we don't want to call threadPaused(), which is
 // expensive.  BUT we must be careful not to violate the invariant
@@ -626,9 +626,7 @@ do_return:
         // the stack, and start executing the BCO.
         INTERP_TICK(it_retto_BCO);
         Sp_subW(1);
-        SpW(0) = (W_)obj;
-        // NB. return the untagged object; the bytecode expects it to
-        // be untagged.  XXX this doesn't seem right.
+        SpW(0) = (W_)tagged_obj;
         obj = (StgClosure*)SpW(2);
         ASSERT(get_itbl(obj)->type == BCO);
         goto run_BCO_return;
@@ -1013,12 +1011,9 @@ run_BCO:
         register StgWord16* instrs    = (StgWord16*)(bco->instrs->payload);
         register StgWord*  literals   = (StgWord*)(&bco->literals->payload[0]);
         register StgPtr*   ptrs       = (StgPtr*)(&bco->ptrs->payload[0]);
-#if defined(DEBUG)
-        int bcoSize;
-        bcoSize = bco->instrs->bytes / sizeof(StgWord16);
-#endif
+        int bcoSize = bco->instrs->bytes / sizeof(StgWord16);
         IF_DEBUG(interpreter,debugBelch("bcoSize = %d\n", bcoSize));
-        IF_DEBUG(interpreter,disassemble( bco ));
+
 #if defined(INTERP_STATS)
         it_lastopc = 0; /* no opcode */
 #endif
@@ -1612,13 +1607,14 @@ run_BCO:
             ASSERT(get_itbl(ap->fun)->type == BCO
                    && BCO_BITMAP_SIZE(ap->fun) == ap->n_args);
 
-            for (i = 0; i < n_payload; i++)
+            for (i = 0; i < n_payload; i++) {
                 ap->payload[i] = (StgClosure*)SpW(i+1);
-                Sp_addW(n_payload+1);
-                IF_DEBUG(interpreter,
-                         debugBelch("\tBuilt ");
-                         printObj((StgClosure*)ap);
-                    );
+            }
+            Sp_addW(n_payload+1);
+            IF_DEBUG(interpreter,
+                     debugBelch("\tBuilt ");
+                     printObj((StgClosure*)ap);
+                );
             goto nextInsn;
         }
 
@@ -1638,13 +1634,14 @@ run_BCO:
                 barf("bci_MKPAP");
             }
 
-            for (i = 0; i < n_payload; i++)
+            for (i = 0; i < n_payload; i++) {
                 pap->payload[i] = (StgClosure*)SpW(i+1);
-                Sp_addW(n_payload+1);
-                IF_DEBUG(interpreter,
-                         debugBelch("\tBuilt ");
-                         printObj((StgClosure*)pap);
-                    );
+            }
+            Sp_addW(n_payload+1);
+            IF_DEBUG(interpreter,
+                     debugBelch("\tBuilt ");
+                     printObj((StgClosure*)pap);
+                );
             goto nextInsn;
         }
 
@@ -1676,10 +1673,11 @@ run_BCO:
             Sp_subW(1);
             // No write barrier is needed here as this is a new allocation
             // visible only from our stack
-            SET_HDR(con, (StgInfoTable*)BCO_LIT(o_itbl), cap->r.rCCCS);
+            StgInfoTable *con_itbl = (StgInfoTable*) BCO_LIT(o_itbl);
+            SET_HDR(con, con_itbl, cap->r.rCCCS);
 
             // Note [Data constructor dynamic tags]
-            //
+            // ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
             // compute the pointer tag for the constructor and tag the pointer
             //
             //     - 1..(TAG_MASK-1): for first TAG_MASK-1 constructors
@@ -1703,7 +1701,7 @@ run_BCO:
         case bci_TESTLT_P: {
             unsigned int discr  = BCO_NEXT;
             int failto = BCO_GET_LARGE_ARG;
-            StgClosure* con = (StgClosure*)SpW(0);
+            StgClosure* con = UNTAG_CLOSURE((StgClosure*)SpW(0));
             if (GET_TAG(con) >= discr) {
                 bciPtr = failto;
             }
@@ -1713,7 +1711,7 @@ run_BCO:
         case bci_TESTEQ_P: {
             unsigned int discr  = BCO_NEXT;
             int failto = BCO_GET_LARGE_ARG;
-            StgClosure* con = (StgClosure*)SpW(0);
+            StgClosure* con = UNTAG_CLOSURE((StgClosure*)SpW(0));
             if (GET_TAG(con) != discr) {
                 bciPtr = failto;
             }
@@ -2001,6 +1999,7 @@ run_BCO:
             // it might have moved during the call.  Also reload the
             // pointers to the components of the BCO.
             obj        = (StgClosure*)SpW(1);
+              // N.B. this is a BCO and therefore is by definition not tagged
             bco        = (StgBCO*)obj;
             instrs     = (StgWord16*)(bco->instrs->payload);
             literals   = (StgWord*)(&bco->literals->payload[0]);
