@@ -18,8 +18,6 @@
 
 void printLoadedObjects(void);
 
-#include "BeginPrivate.h"
-
 /* Which object file format are we targeting? */
 #if defined(linux_HOST_OS) || defined(solaris2_HOST_OS) \
 || defined(linux_android_HOST_OS) \
@@ -38,6 +36,30 @@ typedef char SymbolName;
 typedef struct _ObjectCode ObjectCode;
 typedef struct _Section    Section;
 
+/*
+ * Note [Processing overflowed relocations]
+ * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
+ * When processing relocations whose targets exceed the relocation's maximum
+ * displacement, we can take advantage of knowledge of the symbol type to avoid
+ * linker failures. In particular, if we know that a symbol is a code symbol
+ * then we can handle the relocation by creating a "jump island", a small bit
+ * of code which immediately jumps (with an instruction sequence capable of
+ * larger displacement) to the target.
+ *
+ * This is not possible for data symbols (or, for that matter, Haskell symbols
+ * when TNTC is in use). In these cases we have to rather fail and ask the user
+ * to recompile their program as position-independent.
+ */
+
+/* What kind of thing a symbol identifies. We need to know this to determine how
+ * to process overflowing relocations. See Note [Processing overflowed relocations]. */
+typedef enum _SymType {
+    SYM_TYPE_CODE, /* the symbol is a function and can be relocated via a jump island */
+    SYM_TYPE_DATA, /* the symbol is data */
+    SYM_TYPE_INDIRECT_DATA, /* see Note [_iob_func symbol] */
+} SymType;
+
+
 #if defined(OBJFORMAT_ELF)
 #  include "linker/ElfTypes.h"
 #elif defined(OBJFORMAT_PEi386)
@@ -55,6 +77,7 @@ typedef struct _Symbol
 {
     SymbolName *name;
     SymbolAddr *addr;
+    SymType type;
 } Symbol_t;
 
 typedef struct NativeCodeRange_ {
@@ -77,16 +100,26 @@ typedef
           /* Static initializer section. e.g. .ctors.  */
           SECTIONKIND_INIT_ARRAY,
           /* Static finalizer section. e.g. .dtors.  */
-          SECTIONKIND_FINIT_ARRAY,
+          SECTIONKIND_FINI_ARRAY,
           /* We don't know what the section is and don't care.  */
           SECTIONKIND_OTHER,
+
+          /*
+           * Windows-specific section kinds
+           */
+
           /* Section contains debug information. e.g. .debug$.  */
           SECTIONKIND_DEBUG,
+          /* Section contains exception table. e.g. .pdata.  */
+          SECTIONKIND_EXCEPTION_TABLE,
+          /* Section contains unwind info. e.g. .xdata.  */
+          SECTIONKIND_EXCEPTION_UNWIND,
           /* Section belongs to an import section group. e.g. .idata$.  */
           SECTIONKIND_IMPORT,
+          /* Section defines the head section of a BFD-style import library, e.g. idata$7.  */
+          SECTIONKIND_BFD_IMPORT_LIBRARY_HEAD,
           /* Section defines an import library entry, e.g. idata$7.  */
-          SECTIONKIND_IMPORT_LIBRARY,
-          SECTIONKIND_NOINFOAVAIL
+          SECTIONKIND_BFD_IMPORT_LIBRARY,
         }
    SectionKind;
 
@@ -118,6 +151,7 @@ typedef enum {
 
 /*
  * Note [No typedefs for customizable types]
+ * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  * Some pointer-to-struct types are defined opaquely
  * first, and customized later to architecture/ABI-specific
  * instantiations. Having the usual
@@ -205,6 +239,8 @@ typedef enum {
     DYNAMIC_OBJECT,
 } ObjectType;
 
+typedef void (*cxa_finalize_fn)(void *);
+
 /* Top-level structure for an object module.  One of these is allocated
  * for each object file in use.
  */
@@ -242,6 +278,11 @@ struct _ObjectCode {
        after allocation, so that we can use realloc */
     int        misalignment;
 
+    /* The address of __cxa_finalize; set when at least one finalizer was
+     * register and therefore we must call __cxa_finalize before unloading.
+     * See Note [Resolving __dso_handle]. */
+    cxa_finalize_fn cxa_finalize;
+
     /* The section-kind entries for this object module. An array. */
     int n_sections;
     Section* sections;
@@ -286,12 +327,6 @@ struct _ObjectCode {
        outside one of these is an error in the linker. */
     ProddableBlock* proddables;
 
-#if defined(ia64_HOST_ARCH)
-    /* Procedure Linkage Table for this object */
-    void *plt;
-    unsigned int pltIndex;
-#endif
-
 #if defined(NEED_SYMBOL_EXTRAS)
     SymbolExtra    *symbol_extras;
     unsigned long   first_symbol_extra;
@@ -333,6 +368,12 @@ struct _ObjectCode {
       (OC)->fileName                            \
     )
 
+#define ocDebugBelch(oc, s, ...) \
+    debugBelch("%s(%" PATH_FMT ": " s, \
+               __func__, \
+               OC_INFORMATIVE_FILENAME(oc), \
+               ##__VA_ARGS__)
+
 
 #if defined(THREADED_RTS)
 extern Mutex linker_mutex;
@@ -340,11 +381,14 @@ extern Mutex linker_mutex;
 #if defined(OBJFORMAT_ELF) || defined(OBJFORMAT_MACHO)
 extern Mutex dl_mutex;
 #endif
-#endif
+#endif /* THREADED_RTS */
 
-/* Type of the initializer */
+/* Type of an initializer */
 typedef void (*init_t) (int argc, char **argv, char **env);
 
+/* Type of a finalizer */
+typedef void (*fini_t) (void);
+
 typedef enum _SymStrength {
     STRENGTH_NORMAL,
     STRENGTH_WEAK,
@@ -367,8 +411,11 @@ typedef struct _RtsSymbolInfo {
     SymbolAddr* value;
     ObjectCode *owner;
     SymStrength strength;
+    SymType type;
 } RtsSymbolInfo;
 
+#include "BeginPrivate.h"
+
 void exitLinker( void );
 
 void freeObjectCode (ObjectCode *oc);
@@ -391,11 +438,12 @@ int ghciInsertSymbolTable(
     const SymbolName* key,
     SymbolAddr* data,
     SymStrength weak,
+    SymType type,
     ObjectCode *owner);
 
 /* Lock-free version of lookupSymbol. When 'dependent' is not NULL, adds it as a
- * dependent to the owner of the symbol. */
-SymbolAddr* lookupDependentSymbol (SymbolName* lbl, ObjectCode *dependent);
+ * dependent to the owner of the symbol. The type of the symbol is stored in 'type'. */
+SymbolAddr* lookupDependentSymbol (SymbolName* lbl, ObjectCode *dependent, SymType *type);
 
 /* Perform TLSGD symbol lookup returning the address of the resulting GOT entry,
  * which in this case holds the module id and the symbol offset. */
