@@ -6,7 +6,7 @@
  *
  * ---------------------------------------------------------------------------*/
 
-#include "PosixSource.h"
+#include "rts/PosixSource.h"
 #include "Rts.h"
 
 #include "RtsFlags.h"
@@ -69,7 +69,7 @@ static Time *GC_coll_cpu = NULL;
 static Time *GC_coll_elapsed = NULL;
 static Time *GC_coll_max_pause = NULL;
 
-static int statsPrintf( char *s, ... ) GNUC3_ATTRIBUTE(format (PRINTF, 1, 2));
+static int statsPrintf( char *s, ... ) STG_PRINTF_ATTR(1, 2);
 static void statsFlush( void );
 static void statsClose( void );
 
@@ -766,8 +766,7 @@ StgInt TOTAL_CALLS=1;
 
 /*
 Note [RTS Stats Reporting]
-==========================
-
+~~~~~~~~~~~~~~~~~~~~~~~~~~
 There are currently three reporting functions:
   * report_summary:
       Responsible for producing '+RTS -s' output.
@@ -861,15 +860,18 @@ static void report_summary(const RTSSummaryStats* sum)
                     TimeToSecondsDbl(gen_stats->max_pause_ns));
     }
     if (RtsFlags.GcFlags.useNonmoving) {
-        const int n_major_colls = sum->gc_summary_stats[RtsFlags.GcFlags.generations-1].collections;
-        statsPrintf("  Gen  1     %5d syncs"
+        const uint32_t nonmoving_gen = RtsFlags.GcFlags.generations-1;
+        const int n_major_colls = sum->gc_summary_stats[nonmoving_gen].collections;
+        statsPrintf("  Gen %2d     %5d syncs"
                     ",                      %6.3fs     %3.4fs    %3.4fs\n",
+                    nonmoving_gen,
                     n_major_colls,
                     TimeToSecondsDbl(stats.nonmoving_gc_sync_elapsed_ns),
                     TimeToSecondsDbl(stats.nonmoving_gc_sync_elapsed_ns) / n_major_colls,
                     TimeToSecondsDbl(stats.nonmoving_gc_sync_max_elapsed_ns));
-        statsPrintf("  Gen  1      concurrent"
+        statsPrintf("  Gen %2d      concurrent"
                     ",             %6.3fs  %6.3fs     %3.4fs    %3.4fs\n",
+                    nonmoving_gen,
                     TimeToSecondsDbl(stats.nonmoving_gc_cpu_ns),
                     TimeToSecondsDbl(stats.nonmoving_gc_elapsed_ns),
                     TimeToSecondsDbl(stats.nonmoving_gc_elapsed_ns) / n_major_colls,
@@ -890,7 +892,7 @@ static void report_summary(const RTSSummaryStats* sum)
                 "(%d bound, %d peak workers (%d total), using -N%d)\n\n",
                 taskCount, sum->bound_task_count,
                 peakWorkerCount, workerCount,
-                getNumCapabilities());
+                n_capabilities);
 
     statsPrintf("  SPARKS: %" FMT_Word64
                 " (%" FMT_Word " converted, %" FMT_Word " overflowed, %"
@@ -947,7 +949,7 @@ static void report_summary(const RTSSummaryStats* sum)
                 sum->productivity_cpu_percent * 100,
                 sum->productivity_elapsed_percent * 100);
 
-    // See Note [Internal Counter Stats] for a description of the
+    // See Note [Internal Counters Stats] for a description of the
     // following counters. If you add a counter here, please remember
     // to update the Note.
     if (RtsFlags.MiscFlags.internalCounters) {
@@ -1131,7 +1133,7 @@ static void report_machine_readable (const RTSSummaryStats * sum)
     MR_STAT("work_balance", "f", sum->work_balance);
 
     // next, globals (other than internal counters)
-    MR_STAT("n_capabilities", FMT_Word32, getNumCapabilities());
+    MR_STAT("n_capabilities", FMT_Word32, n_capabilities);
     MR_STAT("task_count", FMT_Word32, taskCount);
     MR_STAT("peak_worker_count", FMT_Word32, peakWorkerCount);
     MR_STAT("worker_count", FMT_Word32, workerCount);
@@ -1346,7 +1348,7 @@ stat_exitReport (void)
     #if defined(THREADED_RTS)
             sum.bound_task_count = taskCount - workerCount;
 
-            for (uint32_t i = 0; i < getNumCapabilities(); i++) {
+            for (uint32_t i = 0; i < n_capabilities; i++) {
                 sum.sparks.created   += capabilities[i]->spark_stats.created;
                 sum.sparks.dud       += capabilities[i]->spark_stats.dud;
                 sum.sparks.overflowed+=
@@ -1473,7 +1475,7 @@ void stat_exit()
 }
 
 /* Note [Work Balance]
-----------------------
+~~~~~~~~~~~~~~~~~~~~~~
 Work balance is a measure of how evenly the work done during parallel garbage
 collection is spread across threads. To compute work balance we must take care
 to account for the number of GC threads changing between GCs. The statistics we
@@ -1552,7 +1554,7 @@ See #13830
 
 /*
 Note [Internal Counters Stats]
------------------------------
+~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 What do the counts at the end of a '+RTS -s --internal-counters' report mean?
 They are detailed below. Most of these counters are used by multiple threads
 with no attempt at synchronisation. This means that reported values  may be
@@ -1645,7 +1647,7 @@ statDescribeGens(void)
       gen_blocks = genLiveBlocks(gen);
 
       mut = 0;
-      for (i = 0; i < getNumCapabilities(); i++) {
+      for (i = 0; i < n_capabilities; i++) {
           mut += countOccupied(capabilities[i]->mut_lists[g]);
 
           // Add the pinned object block.
