@@ -3,13 +3,13 @@
  * (c) The GHC Team 1995-2002
  *
  * Support for concurrent non-blocking I/O and thread waiting in the
- * non-threaded RTS.  In the threded RTS, this file is not used at
+ * non-threaded RTS.  In the threaded RTS, this file is not used at
  * all, instead we use the IO manager thread implemented in Haskell in
  * the base package.
  *
  * ---------------------------------------------------------------------------*/
 
-#include "PosixSource.h"
+#include "rts/PosixSource.h"
 #include "Rts.h"
 
 #include "Signals.h"
@@ -106,8 +106,8 @@ static bool wakeUpSleepingThreads (LowResTime now)
         sleeping_queue = tso->_link;
         tso->why_blocked = NotBlocked;
         tso->_link = END_TSO_QUEUE;
-        IF_DEBUG(scheduler, debugBelch("Waking up sleeping thread %lu\n",
-                                       (unsigned long)tso->id));
+        IF_DEBUG(scheduler, debugBelch("Waking up sleeping thread %"
+                                       FMT_StgThreadID "\n", tso->id));
         // MainCapability: this code is !THREADED_RTS
         pushOnRunQueue(&MainCapability,tso);
         flag = true;
@@ -420,15 +420,15 @@ awaitEvent(bool wait)
                    * pass an IOError to blocked threads (#4934)
                    */
                   IF_DEBUG(scheduler,
-                      debugBelch("Killing blocked thread %lu on bad fd=%i\n",
-                                 (unsigned long)tso->id, fd));
+                      debugBelch("Killing blocked thread %" FMT_StgThreadID
+                                 " on bad fd=%i\n", tso->id, fd));
                   raiseAsync(&MainCapability, tso,
                       (StgClosure *)blockedOnBadFD_closure, false, NULL);
                   break;
               case RTS_FD_IS_READY:
                   IF_DEBUG(scheduler,
-                      debugBelch("Waking up blocked thread %lu\n",
-                                 (unsigned long)tso->id));
+                      debugBelch("Waking up blocked thread %" FMT_StgThreadID "\n",
+                                 tso->id));
                   tso->why_blocked = NotBlocked;
                   tso->_link = END_TSO_QUEUE;
                   pushOnRunQueue(&MainCapability,tso);
