@@ -27,10 +27,6 @@
 #include "sm/Storage.h"
 #include "CNF.h"
 
-#if defined(THREADED_RTS)
-static void nonmovingResetUpdRemSetQueue (MarkQueue *rset);
-static void nonmovingResetUpdRemSet (UpdRemSet *rset);
-#endif
 static bool check_in_nonmoving_heap(StgClosure *p);
 static void mark_closure (MarkQueue *queue, const StgClosure *p, StgClosure **origin);
 static void trace_tso (MarkQueue *queue, StgTSO *tso);
@@ -39,9 +35,6 @@ static void trace_PAP_payload (MarkQueue *queue,
                                StgClosure *fun,
                                StgClosure **payload,
                                StgWord size);
-#if defined(DEBUG)
-static bool is_nonmoving_weak(StgWeak *weak);
-#endif
 
 // How many Array# entries to add to the mark queue at once?
 #define MARK_ARRAY_CHUNK_LENGTH 128
@@ -166,7 +159,6 @@ StgIndStatic *debug_caf_list_snapshot = (StgIndStatic*)END_OF_CAF_LIST;
  *
  * Note [Eager update remembered set flushing]
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- *
  * We eagerly flush update remembered sets during minor GCs to avoid scenarios
  * like the following which could result in long sync pauses:
  *
@@ -206,7 +198,6 @@ StgIndStatic *debug_caf_list_snapshot = (StgIndStatic*)END_OF_CAF_LIST;
  *
  * Note [Concurrent read barrier on deRefWeak#]
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
- *
  * In general the non-moving GC assumes that all pointers reachable from a
  * marked object are themselves marked (or in the mark queue). However,
  * weak pointers are an obvious exception to this rule. In particular,
@@ -257,7 +248,7 @@ StgWord nonmoving_write_barrier_enabled = false;
 MarkQueue *current_mark_queue = NULL;
 
 /* Initialise update remembered set data structures */
-void nonmovingMarkInit() {
+void nonmovingMarkInitUpdRemSet() {
 #if defined(THREADED_RTS)
     initMutex(&upd_rem_set_lock);
     initCondition(&upd_rem_set_flushed_cond);
@@ -270,57 +261,33 @@ static uint32_t markQueueLength(MarkQueue *q);
 #endif
 static void init_mark_queue_(MarkQueue *queue);
 
-static void nonmovingAddUpdRemSetBlocks_(MarkQueue *rset)
+/* Transfers the given capability's update-remembered set to the global
+ * remembered set.
+ *
+ * Really the argument type should be UpdRemSet* but this would be rather
+ * inconvenient without polymorphism.
+ */
+void nonmovingAddUpdRemSetBlocks(MarkQueue *rset)
 {
-    // find the tail of the remembered set mark queue
+    if (markQueueIsEmpty(rset)) return;
+
+    // find the tail of the queue
     bdescr *start = rset->blocks;
     bdescr *end = start;
     while (end->link != NULL)
         end = end->link;
-    rset->blocks = NULL;
 
     // add the blocks to the global remembered set
     ACQUIRE_LOCK(&upd_rem_set_lock);
     end->link = upd_rem_set_block_list;
     upd_rem_set_block_list = start;
     RELEASE_LOCK(&upd_rem_set_lock);
-}
 
-/*
- * Transfers the given capability's update-remembered set to the global
- * remembered set.
- *
- * Really the argument type should be UpdRemSet* but this would be rather
- * inconvenient without polymorphism.
- */
-static void nonmovingAddUpdRemSetBlocks_lock(MarkQueue *rset)
-{
-    if (markQueueIsEmpty(rset)) return;
-
-    nonmovingAddUpdRemSetBlocks_(rset);
-    // Reset the state of the remembered set.
+    // Reset remembered set
     ACQUIRE_SM_LOCK;
     init_mark_queue_(rset);
-    RELEASE_SM_LOCK;
     rset->is_upd_rem_set = true;
-}
-
-/*
- * Transfers the given capability's update-remembered set to the global
- * remembered set.
- *
- * Really the argument type should be UpdRemSet* but this would be rather
- * inconvenient without polymorphism.
- *
- * Caller must hold SM_LOCK.
- */
-void nonmovingAddUpdRemSetBlocks(UpdRemSet *rset)
-{
-    if (markQueueIsEmpty(&rset->queue)) return;
-
-    nonmovingAddUpdRemSetBlocks_(&rset->queue);
-    init_mark_queue_(&rset->queue);
-    rset->queue.is_upd_rem_set = true;
+    RELEASE_SM_LOCK;
 }
 
 #if defined(THREADED_RTS)
@@ -334,7 +301,7 @@ void nonmovingFlushCapUpdRemSetBlocks(Capability *cap)
                "Capability %d flushing update remembered set: %d",
                cap->no, markQueueLength(&cap->upd_rem_set.queue));
     traceConcUpdRemSetFlush(cap);
-    nonmovingAddUpdRemSetBlocks_lock(&cap->upd_rem_set.queue);
+    nonmovingAddUpdRemSetBlocks(&cap->upd_rem_set.queue);
     atomic_inc(&upd_rem_set_flush_count, 1);
     signalCondition(&upd_rem_set_flushed_cond);
     // After this mutation will remain suspended until nonmovingFinishFlush
@@ -356,7 +323,7 @@ void nonmovingBeginFlush(Task *task)
     // task suspended due to a foreign call) in which case our requestSync
     // logic won't have been hit. Make sure that everyone so far has flushed.
     // Ideally we want to mark asynchronously with syncing.
-    for (uint32_t i = 0; i < getNumCapabilities(); i++) {
+    for (uint32_t i = 0; i < n_capabilities; i++) {
         nonmovingFlushCapUpdRemSetBlocks(capabilities[i]);
     }
 }
@@ -368,7 +335,7 @@ bool nonmovingWaitForFlush()
 {
     ACQUIRE_LOCK(&upd_rem_set_lock);
     debugTrace(DEBUG_nonmoving_gc, "Flush count %d", upd_rem_set_flush_count);
-    bool finished = upd_rem_set_flush_count == getNumCapabilities();
+    bool finished = upd_rem_set_flush_count == n_capabilities;
     if (!finished) {
         waitCondition(&upd_rem_set_flushed_cond, &upd_rem_set_lock);
     }
@@ -431,8 +398,8 @@ bool nonmovingWaitForFlush()
 void nonmovingFinishFlush(Task *task)
 {
     // See Note [Unintentional marking in resurrectThreads]
-    for (uint32_t i = 0; i < getNumCapabilities(); i++) {
-        nonmovingResetUpdRemSet(&capabilities[i]->upd_rem_set);
+    for (uint32_t i = 0; i < n_capabilities; i++) {
+        reset_upd_rem_set(&capabilities[i]->upd_rem_set);
     }
     // Also reset upd_rem_set_block_list in case some of the UpdRemSets were
     // filled and we flushed them.
@@ -442,7 +409,7 @@ void nonmovingFinishFlush(Task *task)
     debugTrace(DEBUG_nonmoving_gc, "Finished update remembered set flush...");
     traceConcSyncEnd();
     stat_endNonmovingGcSync();
-    releaseAllCapabilities(getNumCapabilities(), NULL, task);
+    releaseAllCapabilities(n_capabilities, NULL, task);
 }
 #endif
 
@@ -457,8 +424,7 @@ push (MarkQueue *q, const MarkQueueEnt *ent)
     if (q->top->head == MARK_QUEUE_BLOCK_ENTRIES) {
         // Yes, this block is full.
         if (q->is_upd_rem_set) {
-            // Flush the block to the global update remembered set
-            nonmovingAddUpdRemSetBlocks_lock(q);
+            nonmovingAddUpdRemSetBlocks(q);
         } else {
             // allocate a fresh block.
             ACQUIRE_SM_LOCK;
@@ -628,7 +594,7 @@ inline void updateRemembSetPushThunk(Capability *cap, StgThunk *thunk)
  * we update the indirectee to ensure that the thunk's free variables remain
  * visible to the concurrent collector.
  *
- * See Note [Update rememembered set].
+ * See Note [Update remembered set].
  */
 void updateRemembSetPushThunkEager(Capability *cap,
                                    const StgThunkInfoTable *info,
@@ -655,16 +621,6 @@ void updateRemembSetPushThunkEager(Capability *cap,
         }
         break;
     }
-    case THUNK_SELECTOR:
-    {
-        StgSelector *sel = (StgSelector *) thunk;
-        if (check_in_nonmoving_heap(sel->selectee)) {
-            // Don't bother to push origin; it makes the barrier needlessly
-            // expensive with little benefit.
-            push_closure(queue, sel->selectee, NULL);
-        }
-        break;
-    }
     case AP:
     {
         StgAP *ap = (StgAP *) thunk;
@@ -674,11 +630,9 @@ void updateRemembSetPushThunkEager(Capability *cap,
         trace_PAP_payload(queue, ap->fun, ap->payload, ap->n_args);
         break;
     }
-    // We may end up here if a thunk update races with another update.
-    // In this case there is nothing to do as the other thread will have
-    // already pushed the updated thunk's free variables to the update
-    // remembered set.
+    case THUNK_SELECTOR:
     case BLACKHOLE:
+        // TODO: This is right, right?
         break;
     // The selector optimization performed by the nonmoving mark may have
     // overwritten a thunk which we are updating with an indirection.
@@ -733,21 +687,26 @@ STATIC_INLINE bool needs_upd_rem_set_mark(StgClosure *p)
     }
 }
 
+static void finish_upd_rem_set_mark_large(bdescr* bd) {
+    // Someone else may have already marked it.
+    ACQUIRE_LOCK(&nonmoving_large_objects_mutex);
+    if (! (bd->flags & BF_MARKED)) {
+        bd->flags |= BF_MARKED;
+        dbl_link_remove(bd, &nonmoving_large_objects);
+        dbl_link_onto(bd, &nonmoving_marked_large_objects);
+        n_nonmoving_large_blocks -= bd->blocks;
+        n_nonmoving_marked_large_blocks += bd->blocks;
+    }
+    RELEASE_LOCK(&nonmoving_large_objects_mutex);
+}
+
 /* Set the mark bit; only to be called *after* we have fully marked the closure */
 STATIC_INLINE void finish_upd_rem_set_mark(StgClosure *p)
 {
     bdescr *bd = Bdescr((StgPtr) p);
     if (bd->flags & BF_LARGE) {
-        // Someone else may have already marked it.
-        ACQUIRE_LOCK(&nonmoving_large_objects_mutex);
-        if (! (bd->flags & BF_MARKED)) {
-            bd->flags |= BF_MARKED;
-            dbl_link_remove(bd, &nonmoving_large_objects);
-            dbl_link_onto(bd, &nonmoving_marked_large_objects);
-            n_nonmoving_large_blocks -= bd->blocks;
-            n_nonmoving_marked_large_blocks += bd->blocks;
-        }
-        RELEASE_LOCK(&nonmoving_large_objects_mutex);
+        // This function is extracted so that this function can be inline
+        finish_upd_rem_set_mark_large(bd);
     } else {
         struct NonmovingSegment *seg = nonmovingGetSegment((StgPtr) p);
         nonmoving_block_idx block_idx = nonmovingGetBlockIdx((StgPtr) p);
@@ -792,6 +751,13 @@ void updateRemembSetPushStack(Capability *cap, StgStack *stack)
     }
 }
 
+void updateRemembSetPushMessageThrowTo(Capability *cap, MessageThrowTo *m) {
+    updateRemembSetPushClosure(cap, (StgClosure *) m->link);
+    updateRemembSetPushClosure(cap, (StgClosure *) m->source);
+    updateRemembSetPushClosure(cap, (StgClosure *) m->target);
+    updateRemembSetPushClosure(cap, (StgClosure *) m->exception);
+}
+
 /*********************************************************
  * Pushing to the mark queue
  *********************************************************/
@@ -814,7 +780,7 @@ void markQueuePushClosure (MarkQueue *q,
 /* TODO: Do we really never want to specify the origin here? */
 void markQueueAddRoot (MarkQueue* q, StgClosure** root)
 {
-    markQueuePushClosureGC(q, *root);
+    markQueuePushClosure(q, *root, NULL);
 }
 
 /* Push a closure to the mark queue without origin information */
@@ -925,7 +891,6 @@ static MarkQueueEnt markQueuePop (MarkQueue *q)
 static void init_mark_queue_ (MarkQueue *queue)
 {
     bdescr *bd = allocGroup(MARK_QUEUE_BLOCKS);
-    ASSERT(queue->blocks == NULL);
     queue->blocks = bd;
     queue->top = (MarkQueueBlock *) bd->start;
     queue->top->head = 0;
@@ -943,28 +908,20 @@ void initMarkQueue (MarkQueue *queue)
 }
 
 /* Must hold sm_mutex. */
-void nonmovingInitUpdRemSet (UpdRemSet *rset)
+void init_upd_rem_set (UpdRemSet *rset)
 {
     init_mark_queue_(&rset->queue);
     rset->queue.is_upd_rem_set = true;
 }
 
-#if defined(THREADED_RTS)
-static void nonmovingResetUpdRemSetQueue (MarkQueue *rset)
+void reset_upd_rem_set (UpdRemSet *rset)
 {
     // UpdRemSets always have one block for the mark queue. This assertion is to
     // update this code if we change that.
-    ASSERT(rset->is_upd_rem_set);
-    ASSERT(rset->blocks->link == NULL);
-    rset->top->head = 0;
+    ASSERT(rset->queue.blocks->link == NULL);
+    rset->queue.top->head = 0;
 }
 
-static void nonmovingResetUpdRemSet (UpdRemSet *rset)
-{
-    nonmovingResetUpdRemSetQueue(&rset->queue);
-}
-#endif
-
 void freeMarkQueue (MarkQueue *queue)
 {
     freeChain_lock(queue->blocks);
@@ -1310,11 +1267,8 @@ mark_closure (MarkQueue *queue, const StgClosure *p0, StgClosure **origin)
             goto done;
 
         case WHITEHOLE:
-            while (*(StgInfoTable* volatile*) &p->header.info == &stg_WHITEHOLE_info)
-#if defined(PARALLEL_GC)
-                busy_wait_nop()
-#endif
-                ;
+            while (*(StgInfoTable* volatile*) &p->header.info == &stg_WHITEHOLE_info);
+                // busy_wait_nop(); // FIXME
             goto try_again;
 
         default:
@@ -1404,7 +1358,7 @@ mark_closure (MarkQueue *queue, const StgClosure *p0, StgClosure **origin)
     else if (bd->flags & BF_PINNED) {
 #if defined(DEBUG)
         bool found_it = false;
-        for (uint32_t i = 0; i < getNumCapabilities(); ++i) {
+        for (uint32_t i = 0; i < n_capabilities; ++i) {
             if (capabilities[i]->pinned_object_block == bd) {
                 found_it = true;
                 break;
@@ -1522,12 +1476,10 @@ mark_closure (MarkQueue *queue, const StgClosure *p0, StgClosure **origin)
         break;
     }
 
-    case WEAK:
-        ASSERT(is_nonmoving_weak((StgWeak*) p));
-        // fallthrough
     gen_obj:
     case CONSTR:
     case CONSTR_NOCAF:
+    case WEAK:
     case PRIM:
     {
         for (StgWord i = 0; i < info->layout.payload.ptrs; i++) {
@@ -1580,15 +1532,8 @@ mark_closure (MarkQueue *queue, const StgClosure *p0, StgClosure **origin)
     }
 
     case THUNK_SELECTOR:
-    {
-        StgSelector *sel = (StgSelector *) p;
-        // We may be able to evaluate this selector which may render the
-        // selectee unreachable. However, we must mark the selectee regardless
-        // to satisfy the snapshot invariant.
-        PUSH_FIELD(sel, selectee);
-        nonmoving_eval_thunk_selector(queue, sel, origin);
+        nonmoving_eval_thunk_selector(queue, (StgSelector*)p, origin);
         break;
-    }
 
     case AP_STACK: {
         StgAP_STACK *ap = (StgAP_STACK *)p;
@@ -1738,23 +1683,15 @@ done:
  *  b. the nursery has been fully evacuated into the non-moving generation.
  *  c. the mark queue has been seeded with a set of roots.
  *
- * If budget is not UNLIMITED_MARK_BUDGET, then we will mark no more than the
- * indicated number of objects and deduct the work done from the budget.
  */
 GNUC_ATTR_HOT void
-nonmovingMark (MarkBudget* budget, MarkQueue *queue)
+nonmovingMark (MarkQueue *queue)
 {
     traceConcMarkBegin();
     debugTrace(DEBUG_nonmoving_gc, "Starting mark pass");
-    uint64_t count = 0;
+    unsigned int count = 0;
     while (true) {
         count++;
-        if (*budget == 0) {
-            return;
-        } else if (*budget != UNLIMITED_MARK_BUDGET) {
-            *budget -= 1;
-        }
-
         MarkQueueEnt ent = markQueuePop(queue);
 
         switch (nonmovingMarkQueueEntryType(&ent)) {
@@ -1883,66 +1820,20 @@ static bool nonmovingIsNowAlive (StgClosure *p)
 
     bdescr *bd = Bdescr((P_)p);
 
-    const uint16_t flags = bd->flags;
-    if (flags & BF_LARGE) {
-        if (flags & BF_PINNED && !(flags & BF_NONMOVING)) {
-            // In this case we have a pinned object living in a non-full
-            // accumulator block which was not promoted to the nonmoving
-            // generation. Assume that the object is alive.
-            // See #22014.
-            return true;
-        }
+    // All non-static objects in the non-moving heap should be marked as
+    // BF_NONMOVING
+    ASSERT(bd->flags & BF_NONMOVING);
 
-        ASSERT(bd->flags & BF_NONMOVING);
+    if (bd->flags & BF_LARGE) {
         return (bd->flags & BF_NONMOVING_SWEEPING) == 0
                    // the large object wasn't in the snapshot and therefore wasn't marked
             || (bd->flags & BF_MARKED) != 0;
                    // The object was marked
     } else {
-        // All non-static objects in the non-moving heap should be marked as
-        // BF_NONMOVING.
-        ASSERT(bd->flags & BF_NONMOVING);
-
-        struct NonmovingSegment *seg = nonmovingGetSegment((StgPtr) p);
-        StgClosure *snapshot_loc =
-          (StgClosure *) nonmovingSegmentGetBlock(seg, nonmovingSegmentInfo(seg)->next_free_snap);
-        if (p >= snapshot_loc && nonmovingGetClosureMark((StgPtr) p) == 0) {
-            /*
-             * In this case we are looking at a block that wasn't allocated
-             * at the time that the snapshot was taken. As we do not mark such
-             * blocks, we must assume that it is reachable.
-             */
-            return true;
-        } else {
-            return nonmovingClosureMarkedThisCycle((P_)p);
-        }
+        return nonmovingClosureMarkedThisCycle((P_)p);
     }
 }
 
-// Mark all Weak#s on nonmoving_old_weak_ptr_list.
-void nonmovingMarkWeakPtrList (struct MarkQueue_ *queue)
-{
-    ASSERT(nonmoving_weak_ptr_list == NULL);
-    for (StgWeak *w = nonmoving_old_weak_ptr_list; w != NULL; w = w->link) {
-        mark_closure(queue, (StgClosure *) w, NULL);
-    }
-}
-
-// Determine whether a weak pointer object is on one of the nonmoving
-// collector's weak pointer lists. Used for sanity checking.
-#if defined(DEBUG)
-static bool is_nonmoving_weak(StgWeak *weak)
-{
-    for (StgWeak *w = nonmoving_old_weak_ptr_list; w != NULL; w = w->link) {
-        if (w == weak) return true;
-    }
-    for (StgWeak *w = nonmoving_weak_ptr_list; w != NULL; w = w->link) {
-        if (w == weak) return true;
-    }
-    return false;
-}
-#endif
-
 // Non-moving heap variant of `tidyWeakList`
 bool nonmovingTidyWeaks (struct MarkQueue_ *queue)
 {
@@ -1951,9 +1842,6 @@ bool nonmovingTidyWeaks (struct MarkQueue_ *queue)
     StgWeak **last_w = &nonmoving_old_weak_ptr_list;
     StgWeak *next_w;
     for (StgWeak *w = nonmoving_old_weak_ptr_list; w != NULL; w = next_w) {
-        // This should have been marked by nonmovingMarkWeaks
-        ASSERT(nonmovingIsNowAlive((StgClosure *) w));
-
         if (w->header.info == &stg_DEAD_WEAK_info) {
             // finalizeWeak# was called on the weak
             next_w = w->link;
@@ -1964,10 +1852,7 @@ bool nonmovingTidyWeaks (struct MarkQueue_ *queue)
         // Otherwise it's a live weak
         ASSERT(w->header.info == &stg_WEAK_info);
 
-        // See Note [Weak pointer processing and the non-moving GC] in
-        // MarkWeak.c
-        bool key_in_nonmoving = Bdescr((StgPtr) w->key)->flags & BF_NONMOVING;
-        if (!key_in_nonmoving || nonmovingIsNowAlive(w->key)) {
+        if (nonmovingIsNowAlive(w->key)) {
             nonmovingMarkLiveWeak(queue, w);
             did_work = true;
 
@@ -1975,7 +1860,7 @@ bool nonmovingTidyWeaks (struct MarkQueue_ *queue)
             *last_w = w->link;
             next_w = w->link;
 
-            // and put it on nonmoving_weak_ptr_list
+            // and put it on the weak ptr list
             w->link = nonmoving_weak_ptr_list;
             nonmoving_weak_ptr_list = w;
         } else {
@@ -1997,8 +1882,7 @@ void nonmovingMarkDeadWeak (struct MarkQueue_ *queue, StgWeak *w)
 
 void nonmovingMarkLiveWeak (struct MarkQueue_ *queue, StgWeak *w)
 {
-    ASSERT(nonmovingIsNowAlive((StgClosure *) w));
-    ASSERT(nonmovingIsNowAlive((StgClosure *) w->key));
+    ASSERT(nonmovingClosureMarkedThisCycle((P_)w));
     markQueuePushClosure_(queue, w->value);
     markQueuePushClosure_(queue, w->finalizer);
     markQueuePushClosure_(queue, w->cfinalizers);
@@ -2012,9 +1896,9 @@ void nonmovingMarkDeadWeaks (struct MarkQueue_ *queue, StgWeak **dead_weaks)
 {
     StgWeak *next_w;
     for (StgWeak *w = nonmoving_old_weak_ptr_list; w; w = next_w) {
-        ASSERT(!nonmovingIsNowAlive(w->key));
+        ASSERT(!nonmovingClosureMarkedThisCycle((P_)(w->key)));
         nonmovingMarkDeadWeak(queue, w);
-        next_w = w->link;
+        next_w = w ->link;
         w->link = *dead_weaks;
         *dead_weaks = w;
     }
